"""
Code for a class for common interfacing with TESS data, such as downloading, sorting, and manipulating.
"""
import itertools
from io import StringIO
from random import shuffle, Random

import astroquery
import lightkurve

try:
    from enum import StrEnum
except ImportError:
    from backports.strenum import StrEnum

import math
import re
import shutil
import sys
import tempfile
from enum import Enum
from pathlib import Path
from typing import Union, List, Dict
import numpy as np
import pandas as pd
import polars as pl
import requests
from astropy.coordinates import SkyCoord
from astropy.io import fits
from astropy.table import Table
from astroquery.mast import Observations, Catalogs
from astroquery.exceptions import TimeoutError as AstroQueryTimeoutError
from astroquery.vizier import Vizier
from bokeh.io import show
from retrying import retry
from bokeh.plotting import figure as Figure

from ramjet.analysis.light_curve_visualizer import plot_light_curve, create_dual_light_curve_figure

import logging

logger = logging.getLogger('ramjet')


class TessFluxType(Enum):
    """
    An enum to represent the types of available fluxes in TESS two minute data.
    """
    SAP = 'SAP_FLUX'
    PDCSAP = 'PDCSAP_FLUX'


class ColumnName(StrEnum):
    """
    An enum for the names of the columns produced by the data interface class.
    """
    TIC_ID = 'TIC ID'
    SECTOR = 'Sector'


def is_common_mast_connection_error(exception: Exception) -> bool:
    """
    Returns if the passed exception is a common MAST connection error. Made for deciding whether to retry a function.

    :param exception: The exception to check.
    :return: A boolean stating if the exception is a common MAST connection error.
    """
    print(f'Retrying on {exception}...', flush=True)
    # TODO: Rename function, as it includes more than just MAST now.
    return (isinstance(exception, AstroQueryTimeoutError) or
            isinstance(exception, TimeoutError) or
            isinstance(exception, requests.exceptions.ReadTimeout) or
            isinstance(exception, requests.exceptions.ChunkedEncodingError) or
            isinstance(exception, requests.exceptions.HTTPError) or
            isinstance(exception, requests.exceptions.ConnectionError) or
            isinstance(exception, ConnectionResetError) or
            isinstance(exception, lightkurve.search.SearchError) or
            isinstance(exception, astroquery.exceptions.RemoteServiceError))


class NoDataProductsFoundException(Exception):
    """An exception when no data products are found from MAST."""
    pass


def get_variable_data_frame_for_tic_id(tic_id):
    """
    Gets a data frame containing all known variables near a TIC target (including the target if applicable).

    :param tic_id: The TIC target to search for variables near.
    :return: A data frame of the variables near the target (including the target if applicable).
    """
    coordinates = get_target_coordinates(tic_id)
    return get_variable_data_frame_for_coordinates(coordinates)


def add_sector_column_to_single_sector_observations(observations: pd.DataFrame) -> pd.DataFrame:
    """
    Adds a column with the sector the data was taken from.

    :param observations: The table of single-sector observations.
    :return: The table with the added sector column.
    """
    observations[ColumnName.SECTOR] = observations['obs_id'].map(get_sector_from_single_sector_obs_id)
    return observations


def add_tic_id_column_to_single_sector_observations(data_frame: pd.DataFrame) -> pd.DataFrame:
    """
    Adds a column with the TIC ID the row is related to.

    :param data_frame: The data frame of single-sector entries.
    :return: The table with the added TIC ID column.
    """
    data_frame[ColumnName.TIC_ID] = data_frame['obs_id'].map(get_tic_id_from_single_sector_obs_id)
    return data_frame


def print_variables_near_tess_target(tic_id):
    """
    Prints all variable stars near a given TESS target (including the target if applicable).

    :param tic_id: The TIC target to search for variables near.
    """
    variable_data_frame = get_variable_data_frame_for_tic_id(tic_id)
    if variable_data_frame.shape[0] == 0:
        print('No known variables found.')
        return
    print('Variable type abbreviation explanations: http://www.sai.msu.su/gcvs/gcvs/vartype.htm')
    print_data_frame = pd.DataFrame()
    print_data_frame['Variable Type'] = variable_data_frame['VarType'].str.decode('utf-8')
    print_data_frame['Max magnitude'] = variable_data_frame['magMax']
    print_data_frame['Period (days)'] = variable_data_frame['Period']
    print_data_frame.sort_values('Max magnitude', inplace=True)
    print_data_frame.reset_index(drop=True, inplace=True)
    print(print_data_frame)


def get_product_list(observations: pd.DataFrame,
                     mast_input_query_chunk_size: int = 1000) -> pd.DataFrame:
    """
    A wrapper for MAST's `get_product_list`, allowing the use of Pandas DataFrames instead of AstroPy Tables.
    Breaks large queries up to make the communication with MAST smoother.

    :param observations: The data frame of observations to get. Will be converted from DataFrame to Table for query.
    :param mast_input_query_chunk_size: The size of the chunks to process.
    :return: The data frame of the product list. Will be converted from Table to DataFrame for use.
    """
    if observations.shape[0] > 1:
        product_list_chunks = []
        for observations_chunk in np.array_split(observations,
                                                 math.ceil(
                                                     observations.shape[0] / mast_input_query_chunk_size)):
            product_list_chunk = get_product_list_chunk(observations_chunk)
            product_list_chunks.append(product_list_chunk)
        product_list = pd.concat(product_list_chunks, ignore_index=True)
    else:
        product_list = get_product_list_chunk(observations)
    return product_list


def get_all_tess_time_series_observations(tic_id: Union[int, List[int]] = None,
                                          mast_input_query_chunk_size: int = 1000) -> pd.DataFrame:
    """
    Gets all TESS time-series observations, limited to science data product level. Breaks large queries up to make
    the communication with MAST smoother.

    :param tic_id: An optional TIC ID or list of TIC IDs to limit the query to.
    :param mast_input_query_chunk_size: The size of the chunks to process.
    :return: The list of time series observations as rows in a Pandas data frame.
    """
    if tic_id is None or np.isscalar(tic_id):
        observations = get_all_tess_time_series_observations_chunk(tic_id)
    else:
        observations_chunks = []
        for tic_id_list_chunk in np.array_split(tic_id, math.ceil(len(tic_id) / mast_input_query_chunk_size)):
            observations_chunk = get_all_tess_time_series_observations_chunk(tic_id_list_chunk)
            observations_chunks.append(observations_chunk)
        observations = pd.concat(observations_chunks, ignore_index=True)
    return observations


def download_two_minute_cadence_light_curve(tic_id: int, sector: int = None,
                                            save_directory: Union[Path, str] = None) -> Path:
    """
    Downloads a light curve from MAST.

    :param tic_id: The TIC ID of the light curve target to download.
    :param sector: The sector to download. If not specified, downloads first available sector.
    :param save_directory: The directory to save the FITS file to. If not specified, uses the system temporary
                           directory.
    :return: The path to the downloaded file.
    """
    observations = get_all_tess_time_series_observations(tic_id=tic_id)
    single_sector_observations = filter_for_single_sector_observations(observations)
    two_minute_observations = filter_out_twenty_second_cadence_observations(single_sector_observations)
    observations_with_sectors = add_sector_column_to_single_sector_observations(two_minute_observations)
    if sector is not None:
        observations_with_sectors = observations_with_sectors[
            observations_with_sectors[ColumnName.SECTOR] == sector]
    else:
        observations_with_sectors = observations_with_sectors.head(1)
    product_list = get_product_list(observations_with_sectors)
    light_curves_product_list = product_list[product_list['productSubGroupDescription'] == 'LC']
    manifest = download_products(light_curves_product_list, data_directory=Path(tempfile.gettempdir()))
    light_curve_path = Path(manifest['Local Path'].iloc[0])
    if save_directory is not None:
        save_directory = Path(save_directory)
        save_directory.mkdir(parents=True, exist_ok=True)
        new_light_curve_path = save_directory.joinpath(light_curve_path.name)
        shutil.move(str(light_curve_path), str(new_light_curve_path))
        light_curve_path = new_light_curve_path
    return light_curve_path


def get_sectors_target_appears_in(tic_id: int) -> List:
    """
    Gets the list of sectors a TESS target appears in.

    :param tic_id: The TIC ID of the target.
    :return: The list of sectors.
    """
    time_series_observations = get_all_tess_time_series_observations(tic_id)
    single_sector_observations = filter_for_single_sector_observations(time_series_observations)
    two_minute_observations = filter_out_twenty_second_cadence_observations(single_sector_observations)
    single_sector_observations = add_sector_column_to_single_sector_observations(two_minute_observations)
    return sorted(single_sector_observations[ColumnName.SECTOR].unique())


def get_all_two_minute_single_sector_observations(tic_ids: List[int] = None) -> pd.DataFrame:
    """
    Gets the data frame containing all the single sector observations, with TIC ID and sector columns.

    :param tic_ids: The TIC IDs to get observations for.
    :return: The data frame containing the observation information.
    """
    tess_observations = get_all_tess_time_series_observations(tic_id=tic_ids)
    single_sector_observations = filter_for_single_sector_observations(tess_observations)
    two_minute_observations = filter_out_twenty_second_cadence_observations(single_sector_observations)
    two_minute_observations = add_tic_id_column_to_single_sector_observations(
        two_minute_observations)
    two_minute_observations = add_sector_column_to_single_sector_observations(
        two_minute_observations)
    return two_minute_observations


def verify_light_curve(light_curve_path: Path):
    """
    The light_curve is checked if it's malformed, and if it is, it is re-downloaded.

    :param light_curve_path: The path of the light_curve.
    """
    try:
        hdu_list = fits.open(str(light_curve_path))
        light_curve = hdu_list[1].data
        _ = light_curve['TIME'][0]  # Basic check if the light_curve file is malformed.
    except (OSError, TypeError):
        print(f'{light_curve_path} seems to be malformed. Re-downloading and replacing.')
        sector = get_sector_from_single_sector_obs_id(str(light_curve_path.stem))
        tic_id = get_tic_id_from_single_sector_obs_id(str(light_curve_path.stem))
        download_two_minute_cadence_light_curve(tic_id, sector, save_directory=light_curve_path.parent)


def download_two_minute_cadence_light_curves(save_directory: Path, limit: Union[None, int] = None):
    """
    Downloads all two minute cadence light curves from TESS.

    :param save_directory: The directory to save the light curves to.
    :param limit: Limits the number of light curves downloaded. Default of None will download all light curves.
    """
    print(f'Starting download of 2-minute cadence light curves to directory `{save_directory}`.')
    save_directory.mkdir(parents=True, exist_ok=True)
    print(f'Retrieving observations list from MAST...')
    single_sector_observations = get_all_two_minute_single_sector_observations()
    print(f'Retrieving data products list from MAST...')
    data_products = get_product_list(single_sector_observations)
    print(f'Downloading light curves...')
    light_curve_data_products = data_products[data_products['productFilename'].str.endswith('lc.fits')]
    if limit is not None:
        light_curve_data_products = light_curve_data_products.sample(frac=1, random_state=0).head(limit)
    download_manifest = download_products(light_curve_data_products, data_directory=save_directory)
    print(f'Moving light curves to {save_directory}...')
    for _, manifest_row in download_manifest.iterrows():
        if manifest_row['Status'] == 'COMPLETE':
            file_path = Path(manifest_row['Local Path'])
            file_path.rename(save_directory.joinpath(file_path.name))
    print('Database ready.')


def load_light_curve_from_fits_file(light_curve_path: Union[str, Path]) -> Dict[str, np.ndarray]:
    """
    Loads a light_curve from a FITS file in a dictionary form with the structure of the FITS arrays.

    :param light_curve_path: The path to the FITS file.
    :return: The light_curve.
    """
    try:
        with fits.open(light_curve_path) as hdu_list:
            light_curve = hdu_list[1].data  # Light curve information is in first extension table.
    except OSError:  # If the FITS file is corrupt, re-download (seems to happen often enough).
        light_curve_path = Path(light_curve_path)  # In case it's currently a string.
        light_curve_path.unlink()
        tic_id, sector = get_tic_id_and_sector_from_file_path(light_curve_path)
        download_two_minute_cadence_light_curve(tic_id=tic_id, sector=sector,
                                                save_directory=light_curve_path.parent)
        with fits.open(light_curve_path) as hdu_list:
            light_curve = hdu_list[1].data  # Light curve information is in first extension table.
    return light_curve


def load_fluxes_and_times_from_fits_file(light_curve_path: Union[str, Path],
                                         flux_type: TessFluxType = TessFluxType.PDCSAP,
                                         remove_nans: bool = True) -> (np.ndarray, np.ndarray):
    """
    Extract the flux and time values from a TESS FITS file.

    :param light_curve_path: The path to the FITS file.
    :param flux_type: The flux type to extract from the FITS file.
    :param remove_nans: Whether or not to remove nans.
    :return: The flux and times values from the FITS file.
    """
    light_curve = load_light_curve_from_fits_file(light_curve_path)
    fluxes = light_curve[flux_type.value]
    times = light_curve['TIME']
    assert times.shape == fluxes.shape
    if remove_nans:
        # noinspection PyUnresolvedReferences
        nan_indexes = np.union1d(np.argwhere(np.isnan(fluxes)), np.argwhere(np.isnan(times)))
        fluxes = np.delete(fluxes, nan_indexes)
        times = np.delete(times, nan_indexes)
    return fluxes, times


def load_fluxes_flux_errors_and_times_from_fits_file(light_curve_path: Union[str, Path],
                                                     flux_type: TessFluxType = TessFluxType.PDCSAP,
                                                     remove_nans: bool = True
                                                     ) -> (np.ndarray, np.ndarray, np.ndarray):
    """
    Extract the flux and time values from a TESS FITS file.

    :param light_curve_path: The path to the FITS file.
    :param flux_type: The flux type to extract from the FITS file.
    :param remove_nans: Whether or not to remove nans.
    :return: The flux and times values from the FITS file.
    """
    light_curve = load_light_curve_from_fits_file(light_curve_path)
    fluxes = light_curve[flux_type.value]
    flux_errors = light_curve[flux_type.value + '_ERR']
    times = light_curve['TIME']
    assert times.shape == fluxes.shape
    if remove_nans:
        # noinspection PyUnresolvedReferences
        nan_indexes = np.union1d(np.argwhere(np.isnan(fluxes)), np.union1d(np.argwhere(np.isnan(times)),
                                                                           np.argwhere(np.isnan(flux_errors))))
        fluxes = np.delete(fluxes, nan_indexes)
        flux_errors = np.delete(flux_errors, nan_indexes)
        times = np.delete(times, nan_indexes)
    return fluxes, flux_errors, times


def plot_light_curve_from_mast(tic_id: int, sector: int = None, exclude_flux_outliers: bool = False,
                               base_data_point_size=3):
    """
    Downloads and plots a light curve from MAST.

    :param tic_id: The TIC ID of the light curve target to download.
    :param sector: The sector to download. If not specified, downloads first available sector.
    :param exclude_flux_outliers: Whether to exclude flux outlier data points when plotting.
    :param base_data_point_size: The size of the data points to use when plotting (and related sizes).
    """
    light_curve_path = download_two_minute_cadence_light_curve(tic_id, sector)
    fluxes, times = load_fluxes_and_times_from_fits_file(light_curve_path)
    if sector is None:
        sector = get_sector_from_single_sector_obs_id(str(light_curve_path.stem))
    title = f'TIC {tic_id} sector {sector}'
    if exclude_flux_outliers:
        title += ' (outliers removed)'
    plot_light_curve(times=times, fluxes=fluxes, title=title, exclude_flux_outliers=exclude_flux_outliers,
                     base_data_point_size=base_data_point_size)


def create_pdcsap_and_sap_comparison_figure_from_mast(tic_id: int, sector: int = None) -> Figure:
    """
    Creates a comparison figure containing both the PDCSAP and SAP signals.

    :param tic_id: The TIC ID of the light curve to plot.
    :param sector: The sector of the light curve to plot.
    :return: The generated figure.
    """
    light_curve_path = download_two_minute_cadence_light_curve(tic_id, sector)
    pdcsap_fluxes, pdcsap_times = load_fluxes_and_times_from_fits_file(light_curve_path, TessFluxType.PDCSAP)
    normalized_pdcsap_fluxes = pdcsap_fluxes / np.median(pdcsap_fluxes)
    sap_fluxes, sap_times = load_fluxes_and_times_from_fits_file(light_curve_path, TessFluxType.SAP)
    normalized_sap_fluxes = sap_fluxes / np.median(sap_fluxes)
    if sector is None:
        _, sector = get_tic_id_and_sector_from_file_path(light_curve_path)
    title = f'TIC {tic_id} sector {sector}'
    figure = create_dual_light_curve_figure(fluxes0=normalized_pdcsap_fluxes, times0=pdcsap_times, name0='PDCSAP',
                                            fluxes1=normalized_sap_fluxes, times1=sap_times, name1='SAP',
                                            title=title, x_axis_label='Time (BTJD)')
    return figure


def show_pdcsap_and_sap_comparison_from_mast(tic_id: int, sector: int = None):
    """
    Shows a comparison figure containing both the PDCSAP and SAP signals.

    :param tic_id: The TIC ID of the light curve to plot.
    :param sector: The sector of the light curve to plot.
    """
    comparison_figure = create_pdcsap_and_sap_comparison_figure_from_mast(tic_id, sector)
    comparison_figure.sizing_mode = 'stretch_width'
    show(comparison_figure)


def show_light_curve(light_curve_path: Path):
    """
    Shows a figure of the light curve at the passed path.

    :param light_curve_path: The path of the light curve.
    """
    fluxes, times = load_fluxes_and_times_from_fits_file(light_curve_path)
    figure = Figure(title=str(light_curve_path), x_axis_label='Flux', y_axis_label='Time', active_drag='box_zoom')
    color = 'mediumblue'
    figure.line(times, fluxes, line_color=color, line_alpha=0.1)
    figure.circle(times, fluxes, line_color=color, line_alpha=0.4, fill_color=color, fill_alpha=0.1)
    figure.sizing_mode = 'stretch_width'
    show(figure)


def get_tic_id_and_sector_from_file_path(file_path: Union[Path, str]):
    """
    Gets the TIC ID and sector from commonly encountered file name patterns.

    :param file_path: The path of the file to extract the TIC ID and sector.
    :return: The TIC ID and sector. The sector might be omitted (as None).
    """
    if isinstance(file_path, str):
        file_path = Path(file_path)
    file_name = file_path.stem
    # Search for the human-readable version. E.g., "TIC 169480782 sector 5"
    match = re.search(r'TIC (\d+) sector (\d+)', file_name)
    if match:
        return int(match.group(1)), int(match.group(2))
    # Search for the TESS obs_id version. E.g., "tess2018319095959-s0005-0000000278956474-0125-s"
    match = re.search(r'tess\d+-s(\d+)-(\d+)-\d+-s', file_name)
    if match:
        return int(match.group(2)), int(match.group(1))
    # Raise an error if none of the patterns matched.
    raise ValueError(f'{file_name} does not match a known pattern to extract TIC ID and sector from.')


def get_variable_data_frame_for_coordinates(coordinates, radius='21s') -> pd.DataFrame:
    """
    Gets a data frame containing all known variables within a radius of the given coordinates.

    :param coordinates: The coordinates to search.
    :param radius: The radius to search. TESS has a pixel size of 21 arcseconds across.
    :return: The data frame of the variables. Returns an empty data frame if none exist.
    """
    variable_table_list = Vizier.query_region(coordinates, radius=radius, catalog='B/gcvs/gcvs_cat')
    if len(variable_table_list) > 0:
        return variable_table_list[0].to_pandas()
    else:
        return pd.DataFrame()


@retry(retry_on_exception=is_common_mast_connection_error, stop_max_attempt_number=10)
def get_tess_input_catalog_row(tic_id: int) -> pd.Series:
    """
    Get the TIC row for a TIC ID.

    :param tic_id: The target's TIC ID.
    :return: The row of the TIC corresponding to the TIC ID.
    """
    initialize_astroquery()
    target_observations = Catalogs.query_criteria(catalog='TIC', ID=tic_id).to_pandas()
    return target_observations.iloc[0]


@retry(retry_on_exception=is_common_mast_connection_error, stop_max_attempt_number=10)
def get_all_tess_time_series_observations_chunk(tic_id: Union[int, List[int]] = None) -> pd.DataFrame:
    """
    Gets all TESS time-series observations, limited to science data product level. Repeats download attempt on
    error.

    :param tic_id: An optional TIC ID or list of TIC IDs to limit the query to.
    :return: The list of time series observations as rows in a Pandas data frame.
    """
    initialize_astroquery()
    if tic_id is None:
        tic_id = []  # When the empty list is passed to `query_criteria`, any value is considered a match.
    tess_observations = Observations.query_criteria(obs_collection='TESS', dataproduct_type='timeseries',
                                                    calib_level=3,  # Science data product level.
                                                    target_name=tic_id)
    return tess_observations.to_pandas()


def get_all_tess_spoc_light_curve_observations(tic_id: Union[int, List[int]],
                                               mast_input_query_chunk_size: int = 1000) -> pd.DataFrame:
    """
    Gets all TESS SPOC light curves. Breaks large queries up to make the communication with MAST smoother.

    :param tic_id: An optional TIC ID or list of TIC IDs to limit the query to.
    :param mast_input_query_chunk_size: The size of the chunks to process.
    :return: The list of time series observations as rows in a Pandas data frame.
    """
    if np.isscalar(tic_id):
        observations = get_all_tess_spoc_light_curve_observations_chunk(tic_id)
    else:
        observations_chunks = []
        for tic_id_list_chunk in np.array_split(tic_id, math.ceil(len(tic_id) / mast_input_query_chunk_size)):
            observations_chunk = get_all_tess_spoc_light_curve_observations_chunk(tic_id_list_chunk)
            observations_chunks.append(observations_chunk)
        observations = pd.concat(observations_chunks, ignore_index=True)
    return observations


@retry(retry_on_exception=is_common_mast_connection_error, stop_max_attempt_number=10)
def get_all_tess_spoc_light_curve_observations_chunk(tic_id: Union[int, List[int]]) -> pd.DataFrame:
    """
    Gets all TESS time-series observations, limited to science data product level. Repeats download attempt on
    error.

    :param tic_id: An optional TIC ID or list of TIC IDs to limit the query to.
    :return: The list of time series observations as rows in a Pandas data frame.
    """
    initialize_astroquery()
    tess_observations = Observations.query_criteria(obs_collection='HLSP', dataproduct_type='timeseries',
                                                    provenance_name='TESS-SPOC',
                                                    calib_level=4,  # Science data product level.
                                                    target_name=tic_id)
    observations_data_frame = tess_observations.to_pandas()
    return observations_data_frame


def get_spoc_tic_id_list_from_mast() -> List[int]:
    sector_data_frames: List[pl.DataFrame] = []
    for sector_index in itertools.count(1):
        response = requests.get(f'https://archive.stsci.edu/hlsps/tess-spoc/target_lists/s{sector_index:04d}.csv')
        if response.status_code != 200:
            break
        csv_string = response.text[1:]  # Remove hashtag from header.
        sector_data_frame = pl.read_csv(StringIO(csv_string))
        sector_data_frames.append(sector_data_frame)
    if len(sector_data_frames) == 0:
        raise ValueError('No SPOC target lists found. Check network connection.')
    target_list_data_frame = pl.concat(sector_data_frames)
    tic_ids = target_list_data_frame['TIC_ID'].to_list()
    return tic_ids


def download_spoc_light_curves_for_tic_ids(tic_ids: List[int], download_directory: Path,
                                           sectors: List[int] | None = None,
                                           limit: int | None = None, chunk_size: int = 1000) -> List[Path]:
    random = Random(0)
    random.shuffle(tic_ids)
    paths = []
    logger.info(f'Downloading light curves to {download_directory}...')
    for tic_id_list_chunk in np.array_split(tic_ids, math.ceil(len(tic_ids) / chunk_size)):
        chunk_limit = limit - len(paths)
        paths_chunk = download_spoc_light_curves_for_tic_ids_chunk(tic_id_list_chunk, download_directory,
                                                                   sectors=sectors,
                                                                   limit=chunk_limit)
        paths.extend(paths_chunk)
        if len(paths) >= limit:
            break
        logger.info(f'{len(paths)} downloaded...')
    logger.info(f'Finished downloading {len(paths)} light curves.')
    return paths


def get_sector_from_spoc_obs_id(obs_id: str) -> int:
    match = re.search(r'hlsp_tess-spoc_tess_phot_\d+-s(\d+)_tess_v1', obs_id)
    if match:
        return int(match.group(1))


def download_spoc_light_curves_for_tic_ids_chunk(tic_ids: List[int], download_directory: Path,
                                                 sectors: List[int] | None = None,
                                                 limit: int | None = None) -> List[Path]:
    light_curve_observations = get_all_tess_spoc_light_curve_observations(tic_id=tic_ids)
    light_curve_observations[ColumnName.SECTOR] = light_curve_observations['obs_id'].map(get_sector_from_spoc_obs_id)
    if sectors is not None:
        light_curve_observations = light_curve_observations[
            light_curve_observations[ColumnName.SECTOR].isin(sectors)]
    data_product_list = get_product_list(light_curve_observations)
    light_curve_data_products = data_product_list[
        data_product_list['productFilename'].str.endswith('lc.fits')]
    if limit is not None:
        light_curve_data_products = light_curve_data_products.sample(frac=1, random_state=0).head(limit)
    light_curve_data_product_download_manifest = download_products(light_curve_data_products,
                                                                   data_directory=download_directory)
    light_curve_paths = []
    for _, manifest_row in light_curve_data_product_download_manifest.iterrows():
        if manifest_row['Status'] == 'COMPLETE':
            original_file_path = Path(manifest_row['Local Path'])
            updated_file_path = download_directory.joinpath(original_file_path.name)
            original_file_path.rename(updated_file_path)
            light_curve_paths.append(updated_file_path)
        else:
            raise ValueError(
                f'{manifest_row["Local Path"]} was found to not have been completed in the download manifest.')
    return light_curve_paths


def initialize_astroquery():
    Observations.TIMEOUT = 2000
    Observations.PAGESIZE = 3000
    Catalogs.TIMEOUT = 2000
    Catalogs.PAGESIZE = 3000
    try:  # Temporary fix for astroquery's update of timeout and pagesize locations.
        Observations._portal_api_connection.TIMEOUT = 2000
        Observations._portal_api_connection.PAGESIZE = 3000
        Catalogs._portal_api_connection.TIMEOUT = 2000
        Catalogs._portal_api_connection.PAGESIZE = 3000
    except AttributeError:
        pass


@retry(retry_on_exception=is_common_mast_connection_error, stop_max_attempt_number=10)
def get_product_list_chunk(observations: pd.DataFrame) -> pd.DataFrame:
    """
    A wrapper for MAST's `get_product_list`, allowing the use of Pandas DataFrames instead of AstroPy Tables.
    Retries on error when communicating with the MAST server.

    :param observations: The data frame of observations to get. Will be converted from DataFrame to Table for query.
    :return: The data frame of the product list. Will be converted from Table to DataFrame for use.
    """
    data_products = Observations.get_product_list(Table.from_pandas(observations))
    return data_products.to_pandas()


@retry(retry_on_exception=is_common_mast_connection_error, stop_max_attempt_number=10)
def download_products(data_products: pd.DataFrame, data_directory: Path) -> pd.DataFrame:
    """
     A wrapper for MAST's `download_products`, allowing the use of Pandas DataFrames instead of AstroPy Tables.
    Retries on error when communicating with the MAST server.

    :param data_products: The data frame of data products to download. Will be converted from DataFrame to Table
                          for sending the request to MAST.
    :param data_directory: The path to download the data to.
    :return: The manifest of the download. Will be converted from Table to DataFrame for use.
    """
    initialize_astroquery()
    manifest = Observations.download_products(Table.from_pandas(data_products), download_dir=str(data_directory))
    if manifest is None:
        raise NoDataProductsFoundException
    return manifest.to_pandas()


def filter_for_single_sector_observations(time_series_observations: pd.DataFrame) -> pd.DataFrame:
    """
    Filters a data frame of observations to get only the single sector observations.

    :param time_series_observations: A data frame of observations to filter for single sector observations.
    :return: The data frame of single sector observations.
    """
    single_sector_observations = time_series_observations[
        time_series_observations['dataURL'].str.endswith('lc.fits')
    ]
    return single_sector_observations.copy()


def filter_out_twenty_second_cadence_observations(time_series_observations: pd.DataFrame) -> pd.DataFrame:
    """
    Removes 20-second cadence data from the observation data frame.

    :param time_series_observations: A data frame of observations to filtered.
    :return: The data frame without 20-second cadence data.
    """
    observations_without_twenty_second_cadence = time_series_observations[
        ~time_series_observations['dataURL'].str.endswith('fast-lc.fits')
    ]
    return observations_without_twenty_second_cadence.copy()


def filter_for_multi_sector_observations(time_series_observations: pd.DataFrame) -> pd.DataFrame:
    """
    Filters a data frame of observations to get only the multi sector observations.

    :param time_series_observations: A data frame of observations to filter for multi sector observations.
    :return: The data frame of multi sector observations.
    """
    multi_sector_observations = time_series_observations[
        time_series_observations['dataURL'].str.endswith('dvt.fits')
    ]
    return multi_sector_observations.copy()


def get_tic_id_from_single_sector_obs_id(obs_id: str) -> int:
    """
    Extracts the TIC ID from a single-sector obs_id string.

    :param obs_id: The obs_id to extract from.
    :return: The extracted TIC ID.
    """
    return int(obs_id.split('-')[2].lstrip('0'))


def get_sector_from_single_sector_obs_id(obs_id: str) -> int:
    """
    Extracts the sector from a single-sector obs_id string.

    :param obs_id: The obs_id to extract from.
    :return: The extracted sector number.
    """
    return int(obs_id.split('-')[1][1:])


def get_target_coordinates(tic_id: int) -> SkyCoord:
    """
    Get the sky coordinates of the target by a TIC ID.

    :param tic_id: The target's TIC ID.
    :return: The coordinates of the target.
    """
    initialize_astroquery()
    target_observations = Catalogs.query_criteria(catalog='TIC', ID=tic_id).to_pandas()
    ra = target_observations['ra'].iloc[0]
    dec = target_observations['dec'].iloc[0]
    return SkyCoord(ra, dec, unit='deg')


if __name__ == '__main__':
    if len(sys.argv) > 1:
        limit_ = int(sys.argv[1])
    else:
        limit_ = None
    download_two_minute_cadence_light_curves(Path('data/tess_two_minute_cadence_light_curves'), limit=limit_)



import math
from enum import Enum

import numpy as np
import scipy.stats
from numpy import typing as npt
from scipy.interpolate import interp1d


class OutOfBoundsInjectionHandlingMethod(Enum):
    """
    An enum of approaches for handling cases where the injectable signal is shorter than the injectee signal.
    """
    ERROR = 'error'
    REPEAT_SIGNAL = 'repeat_signal'
    RANDOM_INJECTION_LOCATION = 'random_inject_location'


class BaselineFluxEstimationMethod(Enum):
    """
    An enum of to designate the type of baseline flux estimation method to use during training.
    """
    MEDIAN = 'median'
    MEDIAN_ABSOLUTE_DEVIATION = 'median_absolute_deviation'


def inject_signal_into_light_curve_with_intermediates(
        light_curve_times: npt.NDArray[np.float64],
        light_curve_fluxes: npt.NDArray[np.float64],
        signal_times: npt.NDArray[np.float64],
        signal_magnifications: npt.NDArray[np.float64],
        out_of_bounds_injection_handling_method: OutOfBoundsInjectionHandlingMethod =
        OutOfBoundsInjectionHandlingMethod.ERROR,
        baseline_flux_estimation_method: BaselineFluxEstimationMethod = BaselineFluxEstimationMethod.MEDIAN
) -> (npt.NDArray[np.float64], npt.NDArray[np.float64], npt.NDArray[np.float64]):
    """
    Injects a synthetic magnification signal into real light curve fluxes.

    :param light_curve_times: The times of the flux observations of the light curve.
    :param light_curve_fluxes: The fluxes of the light curve to be injected into.
    :param signal_times: The times of the synthetic magnifications.
    :param signal_magnifications: The synthetic magnifications to inject.
    :param out_of_bounds_injection_handling_method: The method to use to handle out of bounds injection.
    :param baseline_flux_estimation_method: The method to use to estimate the baseline flux of the light curve
                                            for scaling the signal magnifications.
    :return: The fluxes with the injected signal, the offset signal times, and the signal flux.
    """
    minimum_light_curve_time = np.min(light_curve_times)
    relative_light_curve_times = light_curve_times - minimum_light_curve_time
    relative_signal_times = signal_times - np.min(signal_times)
    signal_time_length = np.max(relative_signal_times)
    light_curve_time_length = np.max(relative_light_curve_times)
    time_length_difference = light_curve_time_length - signal_time_length
    signal_start_offset = (np.random.random() * time_length_difference) + minimum_light_curve_time
    offset_signal_times = relative_signal_times + signal_start_offset
    if baseline_flux_estimation_method == BaselineFluxEstimationMethod.MEDIAN_ABSOLUTE_DEVIATION:
        baseline_flux = scipy.stats.median_abs_deviation(light_curve_fluxes)
        baseline_to_median_absolute_deviation_ratio = 10  # Arbitrarily chosen to give a reasonable scale.
        baseline_flux *= baseline_to_median_absolute_deviation_ratio
    else:
        baseline_flux = np.median(light_curve_fluxes)
    signal_fluxes = (signal_magnifications * baseline_flux) - baseline_flux
    if out_of_bounds_injection_handling_method is OutOfBoundsInjectionHandlingMethod.RANDOM_INJECTION_LOCATION:
        signal_flux_interpolator = interp1d(offset_signal_times, signal_fluxes, bounds_error=False, fill_value=0)
    elif (out_of_bounds_injection_handling_method is OutOfBoundsInjectionHandlingMethod.REPEAT_SIGNAL and
          time_length_difference > 0):
        before_signal_gap = signal_start_offset - minimum_light_curve_time
        after_signal_gap = time_length_difference - before_signal_gap
        minimum_signal_time_step = np.min(np.diff(offset_signal_times))
        before_repeats_needed = math.ceil(before_signal_gap / (signal_time_length + minimum_signal_time_step))
        after_repeats_needed = math.ceil(after_signal_gap / (signal_time_length + minimum_signal_time_step))
        repeated_signal_fluxes = np.tile(signal_fluxes, before_repeats_needed + 1 + after_repeats_needed)
        repeated_signal_times = None
        for repeat_index in range(-before_repeats_needed, after_repeats_needed + 1):
            repeat_signal_start_offset = (signal_time_length + minimum_signal_time_step) * repeat_index
            if repeated_signal_times is None:
                repeated_signal_times = offset_signal_times + repeat_signal_start_offset
            else:
                repeat_index_signal_times = offset_signal_times + repeat_signal_start_offset
                repeated_signal_times = np.concatenate([repeated_signal_times, repeat_index_signal_times])
        signal_flux_interpolator = interp1d(repeated_signal_times, repeated_signal_fluxes, bounds_error=True)
    else:
        signal_flux_interpolator = interp1d(offset_signal_times, signal_fluxes, bounds_error=True)
    interpolated_signal_fluxes = signal_flux_interpolator(light_curve_times)
    fluxes_with_injected_signal = light_curve_fluxes + interpolated_signal_fluxes
    return fluxes_with_injected_signal, offset_signal_times, signal_fluxes



import re

from peewee import Select

from ramjet.data_interface.tess_ffi_light_curve_metadata_manager import TessFfiLightCurveMetadata
from ramjet.photometric_database.derived.tess_ffi_light_curve_collection import TessFfiLightCurveCollection

try:
    # be ready for 3.10 when it drops
    from enum import StrEnum
except ImportError:
    from backports.strenum import StrEnum
from pathlib import Path
from typing import Iterable, Union, List

import numpy as np
import pandas as pd

from ramjet.photometric_database.light_curve_collection import LightCurveCollection


class ColumnName(StrEnum):
    TIME__DAYS = 'time__days'
    MAGNIFICATION = 'magnification'


class SiddhantSolankiHeartBeatSyntheticSignalsCollection(LightCurveCollection):
    def __init__(self):
        super().__init__()
        self.data_directory: Path = Path('data/siddhant_solanki_synthetic_signals')
        self.label = 1

    def get_paths(self) -> Iterable[Path]:
        all_synthetic_signal_paths = self.data_directory.glob('*.txt')
        heart_beat_synthetic_signals = [path for path in all_synthetic_signal_paths
                                            if re.match(r'generated_lc_\d+.txt', path.name) is not None]
        return heart_beat_synthetic_signals

    def load_times_and_magnifications_from_path(self, path: Path) -> (np.ndarray, np.ndarray):
        synthetic_signal_data_frame = pd.read_csv(path, names=[ColumnName.MAGNIFICATION],
                                                  skipinitialspace=True, delim_whitespace=True, skiprows=1)
        synthetic_signal_data_frame.dropna(inplace=True)
        magnifications = synthetic_signal_data_frame[ColumnName.MAGNIFICATION].values
        step_size__days = 0.0069444444
        times = np.arange(0, magnifications.shape[0] * step_size__days, step_size__days)
        assert times.shape[0] == magnifications.shape[0]
        return times, magnifications


class SiddhantSolankiNonHeartBeatSyntheticSignalsCollection(LightCurveCollection):
    def __init__(self):
        super().__init__()
        self.data_directory: Path = Path('data/siddhant_solanki_synthetic_signals')
        self.label = 0

    def get_paths(self) -> Iterable[Path]:
        all_synthetic_signal_paths = self.data_directory.glob('*.txt')
        non_heart_beat_synthetic_signals = [path for path in all_synthetic_signal_paths
                                            if re.match(r'generated_lc_fake_\d+.txt', path.name) is not None]
        return non_heart_beat_synthetic_signals

    def load_times_and_magnifications_from_path(self, path: Path) -> (np.ndarray, np.ndarray):
        synthetic_signal_data_frame = pd.read_csv(path, names=[ColumnName.MAGNIFICATION],
                                                  skipinitialspace=True, delim_whitespace=True, skiprows=1)
        synthetic_signal_data_frame.dropna(inplace=True)
        magnifications = synthetic_signal_data_frame[ColumnName.MAGNIFICATION].values
        step_size__days = 0.0069444444
        times = np.arange(0, magnifications.shape[0] * step_size__days, step_size__days)
        assert times.shape[0] == magnifications.shape[0]
        return times, magnifications

class TessFfiHeartBeatHardNegativeLightcurveCollection(TessFfiLightCurveCollection):
    """
    A class representing the collection of TESS two minute cadence lightcurves containing eclipsing binaries.
    """
    def __init__(self, dataset_splits: Union[List[int], None] = None,
                 magnitude_range: (Union[float, None], Union[float, None]) = (None, None)):
        super().__init__(dataset_splits=dataset_splits, magnitude_range=magnitude_range)
        self.label = 0
        self.hard_negative_ids = list(pd.read_csv('data/heart_beat_hard_negatives.csv')['tic_id'].values)

    def get_sql_query(self) -> Select:
        """
        Gets the SQL query for the database models for the lightcurve collection.
        :return: The SQL query.
        """
        query = super().get_sql_query()
        query = query.where(TessFfiLightCurveMetadata.tic_id.in_(self.hard_negative_ids))
        return query



import warnings
from enum import Enum
from pathlib import Path
from typing import Union

import pandas as pd
import requests

from ramjet.data_interface.tess_data_interface import get_all_two_minute_single_sector_observations, get_product_list, \
    download_products


class ToiColumns(Enum):
    """
    An enum for the names of the TOI columns for Pandas data frames.
    """
    tic_id = 'TIC ID'
    disposition = 'Disposition'
    transit_epoch__bjd = 'Transit epoch (BJD)'
    transit_period__days = 'Transit period (days)'
    transit_duration = 'Transit duration (hours)'
    sector = 'Sector'


class ExofopDisposition(Enum):
    """
    An enum for the ExoFOP dispositions.
    """
    CONFIRMED_PLANET = 'CP'
    KEPLER_CONFIRMED_PLANET = 'KP'
    PLANET_CANDIDATE = 'PC'
    FALSE_POSITIVE = 'FP'


class TessToiDataInterface:
    """
    A data interface for working with the TESS table of objects of interest.
    """
    def __init__(self, data_directory='data/tess_toi'):
        self.data_directory = Path(data_directory)
        self.data_directory.mkdir(parents=True, exist_ok=True)
        self.toi_dispositions_path = self.data_directory.joinpath('toi_dispositions.csv')
        self.ctoi_dispositions_path = self.data_directory.joinpath('ctoi_dispositions.csv')
        self.light_curves_directory = self.data_directory.joinpath('light_curves')
        self.toi_dispositions_: Union[pd.DataFrame, None] = None
        self.ctoi_dispositions_: Union[pd.DataFrame, None] = None

    @property
    def toi_dispositions(self):
        """
        The TOI dispositions data frame property. Will load as an instance attribute on first access. Updates from
        ExoFOP on first access.

        :return: The TOI dispositions data frame.
        """
        if self.toi_dispositions_ is None:
            try:
                self.update_toi_dispositions_file()
            except requests.exceptions.ConnectionError:
                warnings.warn('Unable to connect to update TOI file. Attempting to use existing file...')
            self.toi_dispositions_ = self.load_toi_dispositions_in_project_format()
        return self.toi_dispositions_

    def update_toi_dispositions_file(self):
        """
        Downloads the latest TOI dispositions file.
        """
        toi_csv_url = 'https://exofop.ipac.caltech.edu/tess/download_toi.php?sort=toi&output=csv'
        response = requests.get(toi_csv_url)
        with self.toi_dispositions_path.open('wb') as csv_file:
            csv_file.write(response.content)

    @property
    def ctoi_dispositions(self):
        """
        The CTOI dispositions data frame property. Will load as an instance attribute on first access. Updates from
        ExoFOP on first access.

        :return: The CTOI dispositions data frame.
        """
        if self.ctoi_dispositions_ is None:
            try:
                self.update_ctoi_dispositions_file()
            except requests.exceptions.ConnectionError:
                warnings.warn('Unable to connect to update TOI file. Attempting to use existing file...')
            self.ctoi_dispositions_ = self.load_ctoi_dispositions_in_project_format()
        return self.ctoi_dispositions_

    def update_ctoi_dispositions_file(self):
        """
        Downloads the latest CTOI dispositions file.
        """
        ctoi_csv_url = 'https://exofop.ipac.caltech.edu/tess/download_ctoi.php?sort=ctoi&output=csv'
        response = requests.get(ctoi_csv_url)
        with self.ctoi_dispositions_path.open('wb') as csv_file:
            csv_file.write(response.content)

    def load_toi_dispositions_in_project_format(self) -> pd.DataFrame:
        """
        Loads the ExoFOP TOI table information from CSV to a data frame using a project consistent naming scheme.

        :return: The data frame of the TOI dispositions table.
        """
        columns_to_use = ['TIC ID', 'TFOPWG Disposition', 'Epoch (BJD)', 'Period (days)', 'Duration (hours)', 'Sectors']
        dispositions = pd.read_csv(self.toi_dispositions_path, usecols=columns_to_use)
        dispositions.rename(columns={'TFOPWG Disposition': ToiColumns.disposition.value,
                                     'Epoch (BJD)': ToiColumns.transit_epoch__bjd.value,
                                     'Period (days)': ToiColumns.transit_period__days.value,
                                     'Duration (hours)': ToiColumns.transit_duration.value,
                                     'Sectors': ToiColumns.sector.value}, inplace=True)
        dispositions[ToiColumns.disposition.value] = dispositions[ToiColumns.disposition.value].fillna('')
        dispositions = dispositions[dispositions[ToiColumns.sector.value].notna()]
        dispositions[ToiColumns.sector.value] = dispositions[ToiColumns.sector.value].str.split(',')
        dispositions = dispositions.explode(ToiColumns.sector.value)
        dispositions[ToiColumns.sector.value] = pd.to_numeric(dispositions[ToiColumns.sector.value]
                                                              ).astype(pd.Int64Dtype())
        return dispositions

    def load_ctoi_dispositions_in_project_format(self) -> pd.DataFrame:
        """
        Loads the ExoFOP CTOI table information from CSV to a data frame using a project consistent naming scheme.

        :return: The data frame of the CTOI dispositions table.
        """
        columns_to_use = ['TIC ID', 'TFOPWG Disposition', 'Transit Epoch (BJD)', 'Period (days)', 'Duration (hrs)']
        dispositions = pd.read_csv(self.ctoi_dispositions_path, usecols=columns_to_use)
        dispositions.rename(columns={'TFOPWG Disposition': ToiColumns.disposition.value,
                                     'Transit Epoch (BJD)': ToiColumns.transit_epoch__bjd.value,
                                     'Period (days)': ToiColumns.transit_period__days.value,
                                     'Duration (hrs)': ToiColumns.transit_duration.value}, inplace=True)
        dispositions[ToiColumns.disposition.value] = dispositions[ToiColumns.disposition.value].fillna('')
        return dispositions

    def retrieve_exofop_toi_and_ctoi_planet_disposition_for_tic_id(self, tic_id: int) -> pd.DataFrame:
        """
        Retrieves the ExoFOP disposition information for a given TIC ID from <https://exofop.ipac.caltech.edu/tess/>`_.

        :param tic_id: The TIC ID to get available data for.
        :return: The disposition data frame.
        """
        toi_dispositions = self.toi_dispositions
        ctoi_dispositions = self.ctoi_dispositions
        toi_and_coi_dispositions = pd.concat([toi_dispositions, ctoi_dispositions], axis=0, ignore_index=True)
        tic_target_dispositions = toi_and_coi_dispositions[toi_and_coi_dispositions['TIC ID'] == tic_id]
        return tic_target_dispositions

    def has_any_exofop_dispositions_for_tic_id(self, tic_id: int) -> bool:
        """
        Returns whether or not any dispositions exist for this TIC ID.

        :param tic_id: The TIC ID to check.
        :return: True if there are dispositions, False if none.
        """
        existing_dispositions = self.retrieve_exofop_toi_and_ctoi_planet_disposition_for_tic_id(tic_id)
        if existing_dispositions.shape[0] != 0:
            return True
        else:
            return False

    def print_exofop_toi_and_ctoi_planet_dispositions_for_tic_target(self, tic_id):
        """
        Prints all ExoFOP disposition information for a given TESS target.

        :param tic_id: The TIC target to for.
        """
        dispositions_data_frame = self.retrieve_exofop_toi_and_ctoi_planet_disposition_for_tic_id(tic_id)
        if dispositions_data_frame.shape[0] == 0:
            print('No known ExoFOP dispositions found.')
            return
        # Use context options to not truncate printed data.
        with pd.option_context('display.max_rows', None, 'display.max_columns', None, 'display.width', None):
            print(dispositions_data_frame)
            
    def download_exofop_toi_light_curves_to_directory(self, directory: Path):
        """
        Downloads the `ExoFOP database <https://exofop.ipac.caltech.edu/tess/view_toi.php>`_ light curve files to the
        given directory.

        :param directory: The directory to download the light curves to. Defaults to the data interface directory.
        """
        print("Downloading ExoFOP TOI disposition CSV...")
        if isinstance(directory, str):
            directory = Path(directory)
        tic_ids = self.toi_dispositions[ToiColumns.tic_id.value].unique()
        print('Downloading TESS observation list...')
        single_sector_observations = get_all_two_minute_single_sector_observations(tic_ids)
        print("Downloading light curves which are confirmed or suspected planets in TOI dispositions...")
        suspected_planet_dispositions = self.toi_dispositions[
            self.toi_dispositions[ToiColumns.disposition.value] != 'FP']
        suspected_planet_observations = pd.merge(single_sector_observations, suspected_planet_dispositions, how='inner',
                                                 on=[ToiColumns.tic_id.value, ToiColumns.sector.value])
        suspected_planet_data_products = get_product_list(suspected_planet_observations)
        suspected_planet_light_curve_data_products = suspected_planet_data_products[
            suspected_planet_data_products['productFilename'].str.endswith('lc.fits')]
        suspected_planet_download_manifest = download_products(
            suspected_planet_light_curve_data_products, data_directory=self.data_directory)
        print(f'Verifying and moving light curves to {directory}...')
        directory.mkdir(parents=True, exist_ok=True)
        for row_index, row in suspected_planet_download_manifest.iterrows():
            if row['Status'] == 'COMPLETE':
                file_path = Path(row['Local Path'])
                file_path.rename(directory.joinpath(file_path.name))


if __name__ == '__main__':
    tess_toi_data_interface = TessToiDataInterface()
    tess_toi_data_interface.download_exofop_toi_light_curves_to_directory(
        Path('data/tess_two_minute_cadence_light_curves'))



from pathlib import Path

import numpy as np

from qusi.light_curve import LightCurve
from qusi.light_curve_collection import LabeledLightCurveCollection, create_constant_label_for_path_function


class ToyLightCurve:
    """
    Simple toy light curves.
    """

    @classmethod
    def flat(cls, value: float = 1) -> LightCurve:
        """
        Creates a flat light curve.
        """
        length = 100
        fluxes = np.full(shape=[length], fill_value=value, dtype=np.float32)
        times = np.arange(length, dtype=np.float32)
        return LightCurve.new(times=times, fluxes=fluxes)

    @classmethod
    def sine_wave(cls, period=50) -> LightCurve:
        """
        Creates a sine wave light curve.
        """
        length = 100
        periods_to_produce = length / period
        fluxes = (np.sin(np.linspace(0, np.pi * periods_to_produce, num=length, endpoint=False)) / 2) + 1
        times = np.arange(length)
        return LightCurve.new(times=times, fluxes=fluxes)


def toy_light_curve_get_paths_function():
    return [Path('')]

def toy_flat_light_curve_load_times_and_fluxes(path: Path) -> (np.ndarray, np.ndarray):
    light_curve = ToyLightCurve.flat()
    return light_curve.times, light_curve.fluxes

def toy_sine_wave_light_curve_load_times_and_fluxes(path: Path) -> (np.ndarray, np.ndarray):
    light_curve = ToyLightCurve.sine_wave()
    return light_curve.times, light_curve.fluxes


toy_flat_light_curve_collection = LabeledLightCurveCollection.new(
    get_paths_function=toy_light_curve_get_paths_function,
    load_times_and_fluxes_from_path_function=toy_flat_light_curve_load_times_and_fluxes,
    load_label_from_path_function=create_constant_label_for_path_function(0)
)

toy_sine_wave_light_curve_collection = LabeledLightCurveCollection.new(
    get_paths_function=toy_light_curve_get_paths_function,
    load_times_and_fluxes_from_path_function=toy_sine_wave_light_curve_load_times_and_fluxes,
    load_label_from_path_function=create_constant_label_for_path_function(1)
)



"""
Code representing the collection of TESS two minute cadence light curves containing eclipsing binaries.
"""
from typing import Union, List

from peewee import Select

from ramjet.data_interface.tess_eclipsing_binary_metadata_manager import TessEclipsingBinaryMetadata
from ramjet.data_interface.tess_ffi_light_curve_metadata_manager import TessFfiLightCurveMetadata
from ramjet.data_interface.tess_transit_metadata_manager import TessTransitMetadata, Disposition as TransitDisposition
from ramjet.photometric_database.derived.tess_ffi_light_curve_collection import TessFfiLightCurveCollection


class TessFfiEclipsingBinaryLightCurveCollection(TessFfiLightCurveCollection):
    """
    A class representing the collection of TESS two minute cadence light curves containing eclipsing binaries.
    """
    def __init__(self, dataset_splits: Union[List[int], None] = None,
                 magnitude_range: (Union[float, None], Union[float, None]) = (None, None)):
        super().__init__(dataset_splits=dataset_splits, magnitude_range=magnitude_range)
        self.label = 1

    def get_sql_query(self) -> Select:
        """
        Gets the SQL query for the database models for the light curve collection.

        :return: The SQL query.
        """
        query = super().get_sql_query()
        eclipsing_binary_tic_id_query = TessEclipsingBinaryMetadata.select(TessEclipsingBinaryMetadata.tic_id)
        query = query.where(TessFfiLightCurveMetadata.tic_id.in_(eclipsing_binary_tic_id_query))
        return query


class TessFfiAntiEclipsingBinaryForTransitLightCurveCollection(TessFfiLightCurveCollection):
    """
    A class representing the collection of TESS two minute cadence light curves flagged as eclipsing binaries which are
    not a suspected transit.
    """
    def __init__(self, dataset_splits: Union[List[int], None] = None,
                 magnitude_range: (Union[float, None], Union[float, None]) = (None, None)):
        super().__init__(dataset_splits=dataset_splits, magnitude_range=magnitude_range)
        self.label = 0

    def get_sql_query(self) -> Select:
        """
        Gets the SQL query for the database models for the light curve collection.

        :return: The SQL query.
        """
        query = super().get_sql_query()
        transit_tic_id_query = TessTransitMetadata.select(TessTransitMetadata.tic_id).where(
            (TessTransitMetadata.disposition == TransitDisposition.CONFIRMED.value) |
            (TessTransitMetadata.disposition == TransitDisposition.CANDIDATE.value))
        eclipsing_binary_tic_id_query = TessEclipsingBinaryMetadata.select(TessEclipsingBinaryMetadata.tic_id).where(
            TessEclipsingBinaryMetadata.tic_id.not_in(transit_tic_id_query))
        query = query.where(TessFfiLightCurveMetadata.tic_id.in_(eclipsing_binary_tic_id_query))
        return query



"""
Code for the the metadatabase.
"""
import datetime
from uuid import UUID, uuid5
from typing import Type

from peewee import Model, SqliteDatabase, DateTimeField

metadatabase = SqliteDatabase('data/metadatabase.sqlite3',
                              pragmas={'journal_mode': 'wal'},
                              check_same_thread=False)
metadatabase_uuid_namespace = UUID('ed5c78c4-d8dd-4525-9633-97beac696cd1')


def convert_class_to_table_name(model_class: Type[Model]):
    """
    Creates the table name based on the model class.

    :param model_class: The class to create the table name for.
    :return: The name of the table.
    """
    model_name = model_class.__name__
    return model_name


class MetadatabaseModel(Model):
    """
    A general model for the metadatabase tables.
    """
    class Meta:
        """The meta information for the metadatabase models."""
        database = metadatabase
        table_function = convert_class_to_table_name
        primary_key = False


def metadatabase_uuid(name: str) -> UUID:
    """
    Generates a reproducible UUID for the metadatabase based on a name string.

    :param name: The string used to produce the UUID.
    :return: The UUID.
    """
    return uuid5(metadatabase_uuid_namespace, name)


def dataset_split_from_uuid(uuid: UUID) -> int:
    """
    Generates a repeatable dataset split from a UUID.

    :param uuid: The UUID to seed with.
    :return: The dataset split.
    """
    return uuid.int % 10



"""
Code to prepare the metadata needed for TESS two minute transit related trials.
"""
from ramjet.data_interface.tess_target_metadata_manager import TessTargetMetadataManger
from ramjet.data_interface.tess_transit_metadata_manager import TessTransitMetadataManager
from ramjet.data_interface.tess_two_minute_cadence_light_curve_metadata_manager import \
    TessTwoMinuteCadenceLightCurveMetadataManger


def build_tables():
    """
    Prepare the metadata tables needed for TESS two minute transit related trials.
    """
    tess_two_minute_cadence_light_curve_metadata_manger = TessTwoMinuteCadenceLightCurveMetadataManger()
    tess_two_minute_cadence_light_curve_metadata_manger.build_table()
    tess_transit_metadata_manager = TessTransitMetadataManager()
    tess_transit_metadata_manager.build_table()
    tess_target_metadata_manger = TessTargetMetadataManger()
    tess_target_metadata_manger.build_table()


if __name__ == '__main__':
    build_tables()



"""
Code for managing the TESS transit metadata.
"""
import sqlite3
import warnings
from typing import List

import pandas as pd
from enum import Enum
from peewee import IntegerField, CharField

from ramjet.data_interface.metadatabase import MetadatabaseModel, metadatabase
from ramjet.data_interface.tess_toi_data_interface import TessToiDataInterface, ToiColumns
from ramjet.database.tess_planet_disposition import TessPlanetDisposition


class Disposition(Enum):
    """
    An enum to represent the possible dispositions of a target.
    """
    CONFIRMED = 'Confirmed'
    CANDIDATE = 'Candidate'
    FALSE_POSITIVE = 'False positive'


class TessTransitMetadata(MetadatabaseModel):
    """
    A model for the TESS target transit metadatabase table.
    """
    tic_id = IntegerField(index=True, unique=True)
    disposition = CharField(index=True, choices=Disposition)

    class Meta:
        """Schema meta data for the model."""
        indexes = (
            (('disposition', 'tic_id'), False),
        )


class TessTransitMetadataManager:
    """
    A class for managing the TESS transit metadata.
    """
    @staticmethod
    def build_table():
        """
        Builds the TESS transit metadata table.
        """
        print('Building TESS transit metadata table...')
        tess_toi_data_interface = TessToiDataInterface()
        toi_dispositions = tess_toi_data_interface.toi_dispositions
        ctoi_dispositions = tess_toi_data_interface.ctoi_dispositions
        toi_filtered_dispositions = toi_dispositions.filter([ToiColumns.tic_id.value, ToiColumns.disposition.value])
        ctoi_filtered_dispositions = ctoi_dispositions.filter([ToiColumns.tic_id.value, ToiColumns.disposition.value])
        all_dispositions = pd.concat([toi_filtered_dispositions, ctoi_filtered_dispositions], ignore_index=True)
        target_grouped_dispositions = all_dispositions.groupby(ToiColumns.tic_id.value)[ToiColumns.disposition.value
                                                                                        ].apply(set)
        row_count = 0
        metadatabase.drop_tables([TessTransitMetadata])
        metadatabase.create_tables([TessTransitMetadata])
        with metadatabase.atomic():
            for tic_id, disposition_set in target_grouped_dispositions.items():
                # As a target can have multiple dispositions, use the most forgiving available disposition.
                if 'KP' in disposition_set or 'CP' in disposition_set:
                    database_disposition = Disposition.CONFIRMED.value
                elif 'PC' in disposition_set or '' in disposition_set or 'APC' in disposition_set:
                    database_disposition = Disposition.CANDIDATE.value
                elif 'FP' in disposition_set or 'FA' in disposition_set:
                    database_disposition = Disposition.FALSE_POSITIVE.value
                else:
                    warnings.warn(f'Dispositions for TIC {tic_id} are {disposition_set}, which does not contain a known'
                                  f'disposition.')
                    continue
                row = TessTransitMetadata(tic_id=tic_id, disposition=database_disposition)
                row.save()
                row_count += 1
        print(f'Table built. {row_count} rows added.')

    @staticmethod
    def add_tic_ids_as_confirmed(tic_ids: List[int]):
        """
        Adds the passed TIC IDs as confirmed transits.

        :param tic_ids: The list of TIC IDs.
        """
        rows_added = 0
        with metadatabase.atomic():
            for tic_id in tic_ids:
                query = TessTransitMetadata.select().where(TessTransitMetadata.tic_id == tic_id)
                if query.exists():
                    continue
                transit = TessTransitMetadata()
                transit.tic_id = tic_id
                transit.disposition = Disposition.CONFIRMED.value
                transit.save()
                rows_added += 1
        print(f'{rows_added} rows added.')


if __name__ == '__main__':
    tess_transit_metadata_manager = TessTransitMetadataManager()
    tess_transit_metadata_manager.build_table()
    try:
        tess_transit_metadata_manager.add_tic_ids_as_confirmed(
            TessPlanetDisposition.get_tic_ids_of_passing_vetted_transiting_planet_candidates())
    except sqlite3.OperationalError:
        pass



"""
An abstract class allowing for any number and combination of standard and injectable/injectee light curve collections.
"""
from functools import partial
from typing import List, Union, Callable, Tuple, Optional

import numpy as np
import numpy.typing as npt

from ramjet.logging.wandb_logger import WandbLogger, WandbLoggableInjection
from ramjet.photometric_database.light_curve import LightCurve
from ramjet.photometric_database.light_curve_collection import LightCurveCollection
from ramjet.photometric_database.light_curve_database import LightCurveDatabase
from ramjet.photometric_database.light_curve_dataset_manipulations import OutOfBoundsInjectionHandlingMethod, \
    BaselineFluxEstimationMethod, inject_signal_into_light_curve_with_intermediates


def inject_signal_into_light_curve(
        light_curve_times: npt.NDArray[np.float64],
        light_curve_fluxes: npt.NDArray[np.float64],
        signal_times: npt.NDArray[np.float64],
        signal_magnifications: npt.NDArray[np.float64],
        out_of_bounds_injection_handling_method: OutOfBoundsInjectionHandlingMethod =
        OutOfBoundsInjectionHandlingMethod.ERROR,
        baseline_flux_estimation_method: BaselineFluxEstimationMethod = BaselineFluxEstimationMethod.MEDIAN
) -> npt.NDArray[np.float64]:
    """
    Injects a synthetic magnification signal into real light curve fluxes.

    :param light_curve_times: The times of the flux observations of the light curve.
    :param light_curve_fluxes: The fluxes of the light curve to be injected into.
    :param signal_times: The times of the synthetic magnifications.
    :param signal_magnifications: The synthetic magnifications to inject.
    :param out_of_bounds_injection_handling_method: The method to use to handle out of bounds injection.
    :param baseline_flux_estimation_method: The method to use to estimate the baseline flux of the light curve
                                            for scaling the signal magnifications.
    :return: The fluxes with the injected signal, the offset signal times, and the signal flux.
    """
    fluxes_with_injected_signal, _, _ = inject_signal_into_light_curve_with_intermediates(
        light_curve_times=light_curve_times,
        light_curve_fluxes=light_curve_fluxes,
        signal_times=signal_times,
        signal_magnifications=signal_magnifications,
        out_of_bounds_injection_handling_method=out_of_bounds_injection_handling_method,
        baseline_flux_estimation_method=baseline_flux_estimation_method
    )
    return fluxes_with_injected_signal


class StandardAndInjectedLightCurveDatabase(LightCurveDatabase):
    """
    An abstract class allowing for any number and combination of standard and injectable/injectee light curve collections
    to be used for training.
    """

    def __init__(self):
        super().__init__()
        self.training_standard_light_curve_collections: List[LightCurveCollection] = []
        self.training_injectee_light_curve_collection: Union[LightCurveCollection, None] = None
        self.training_injectable_light_curve_collections: List[LightCurveCollection] = []
        self.validation_standard_light_curve_collections: List[LightCurveCollection] = []
        self.validation_injectee_light_curve_collection: Union[LightCurveCollection, None] = None
        self.validation_injectable_light_curve_collections: List[LightCurveCollection] = []
        self.inference_light_curve_collections: List[LightCurveCollection] = []
        self.shuffle_buffer_size = 10000
        self.number_of_label_values = 1
        self.number_of_auxiliary_values: int = 0
        self.out_of_bounds_injection_handling: OutOfBoundsInjectionHandlingMethod = \
            OutOfBoundsInjectionHandlingMethod.ERROR
        self.baseline_flux_estimation_method = BaselineFluxEstimationMethod.MEDIAN
        self.logger: Optional[WandbLogger] = None

    @property
    def number_of_input_channels(self) -> int:
        """
        Determines the number of input channels that should exist for this database.

        :return: The number of channels.
        """
        channels = 1
        if self.include_time_as_channel:
            channels += 1
        if self.include_flux_errors_as_channel:
            channels += 1
        return channels

    def add_logging_queues_to_map_function(self, preprocess_map_function: Callable, name: Optional[str]) -> Callable:
        """
        Adds logging queues to the map functions.

        :param preprocess_map_function: The function to map.
        :param name: The name of the dataset.
        :return: The updated map function.
        """
        if self.logger is not None:
            preprocess_map_function = partial(preprocess_map_function,
                                              request_queue=self.logger.create_request_queue_for_collection(name),
                                              example_queue=self.logger.create_example_queue_for_collection(name))
        return preprocess_map_function

    def inject_signal_into_light_curve(self, light_curve_fluxes: np.ndarray, light_curve_times: np.ndarray,
                                       signal_magnifications: np.ndarray, signal_times: np.ndarray,
                                       wandb_loggable_injection: Optional[WandbLoggableInjection] = None) -> np.ndarray:
        """
        Injects a synthetic magnification signal into real light curve fluxes.

        :param light_curve_fluxes: The fluxes of the light curve to be injected into.
        :param light_curve_times: The times of the flux observations of the light curve.
        :param signal_magnifications: The synthetic magnifications to inject.
        :param signal_times: The times of the synthetic magnifications.
        :param wandb_loggable_injection: The object to log the injection process.
        :return: The fluxes with the injected signal.
        """
        out_of_bounds_injection_handling_method = self.out_of_bounds_injection_handling
        baseline_flux_estimation_method = self.baseline_flux_estimation_method
        fluxes_with_injected_signal, offset_signal_times, signal_fluxes = inject_signal_into_light_curve_with_intermediates(
            light_curve_times, light_curve_fluxes, signal_times, signal_magnifications,
            out_of_bounds_injection_handling_method, baseline_flux_estimation_method)
        if wandb_loggable_injection is not None:
            wandb_loggable_injection.aligned_injectee_light_curve = LightCurve.from_times_and_fluxes(
                light_curve_times, light_curve_fluxes)
            wandb_loggable_injection.aligned_injectable_light_curve = LightCurve.from_times_and_fluxes(
                offset_signal_times, signal_fluxes)
            wandb_loggable_injection.aligned_injected_light_curve = LightCurve.from_times_and_fluxes(
                light_curve_times, fluxes_with_injected_signal)
        return fluxes_with_injected_signal


def expand_label_to_training_dimensions(label: Union[int, List[int], Tuple[int], np.ndarray]) -> np.ndarray:
    """
    Expand the label to the appropriate dimensions for training.

    :param label: The label to convert.
    :return: The label with the correct dimensions.
    """
    if type(label) is not np.ndarray:
        if type(label) in [list, tuple]:
            label = np.array(label)
        else:
            label = np.array([label])
    return label



from copy import deepcopy
from dataclasses import dataclass
from typing import Self

from qusi.light_curve import LightCurve, remove_nan_flux_data_points_from_light_curve


@dataclass
class LightCurveObservation:
    """
    An observation containing a light curve and label. Note, this is an observation in machine learning terms, not to be
    confused with an astrophysical observation.

    :ivar light_curve: The light curve.
    :ivar label: The integer classification label.
    """
    light_curve: LightCurve
    label: int

    @classmethod
    def new(cls, light_curve: LightCurve, label: int) -> Self:
        """
        Creates a new LightCurveObservation.

        :param light_curve: The light curve.
        :param label: The integer classification label.
        :return: The observation.
        """
        return cls(light_curve=light_curve, label=label)


def remove_nan_flux_data_points_from_light_curve_observation(
        light_curve_observation: LightCurveObservation) -> LightCurveObservation:
    light_curve_observation = deepcopy(light_curve_observation)
    light_curve_observation.light_curve = remove_nan_flux_data_points_from_light_curve(
        light_curve_observation.light_curve)
    return light_curve_observation



"""
Code for a light curve collection of the MOA data.
"""
from pathlib import Path
from typing import Iterable
import numpy as np
import pandas as pd

from ramjet.photometric_database.light_curve_collection import LightCurveCollection
from ramjet.photometric_database.microlensing_signal_generator import MagnificationSignal
import random


class MOAPositiveMicrolensingLightCurveCollection(LightCurveCollection):
    """
    A light curve collection of the MOA data with previously detected microlensing events.
    """

    def __init__(self, dataset_splits=None, split_pieces=5):
        super().__init__()
        self.label = 1
        self.split_pieces = split_pieces
        if dataset_splits is None:
            dataset_splits = [1, 2, 3, 4]
        self.dataset_splits = dataset_splits

    def get_paths(self) -> Iterable[Path]:
        """
        Gets the paths for the light curves in the positive collection.

        :return: An iterable of the light curve paths.
        """
        paths = Path('/local/data/fugu3/sishitan/ramjet/data/moa_microlensing/positive').glob('*.feather')
        path_list = list(paths)
        random.seed(42)
        random.shuffle(path_list)
        number_of_samples = len(path_list)
        number_of_samples_per_block = number_of_samples // self.split_pieces
        dataset_paths = []
        for block in self.dataset_splits:
            if block == (self.split_pieces - 1):
                dataset_paths += path_list[
                                 block * number_of_samples_per_block:]
            else:
                dataset_paths += path_list[
                                 block * number_of_samples_per_block:(block + 1) * number_of_samples_per_block]
        return dataset_paths

    def load_times_and_fluxes_from_path(self, path: Path) -> (np.ndarray, np.ndarray):
        """
        Loads the times and fluxes from a given light curve path.

        :param path: The path to the light curve file.
        :return: The times and the fluxes of the light curve.
        """
        light_curve_dataframe = pd.read_feather(path)
        times = light_curve_dataframe['HJD'].values
        fluxes = light_curve_dataframe['flux'].values
        return times, fluxes


class MOANegativeMicrolensingLightCurveCollection(LightCurveCollection):
    """
    A light curve collection of the MOA data with no microlensing event.
    """

    def __init__(self, dataset_splits=None, split_pieces=5):
        super().__init__()
        self.label = 0
        self.split_pieces = split_pieces
        if dataset_splits is None:
            dataset_splits = [1, 2, 3, 4]
        self.dataset_splits = dataset_splits

    def get_paths(self):
        """
        Gets the paths for the light curves in the negative collection.

        :return: An iterable of the light curve paths.
        """
        paths = Path('/local/data/fugu3/sishitan/ramjet/data/moa_microlensing/negative').glob('*.feather')
        path_list = list(paths)
        random.seed(42)
        random.shuffle(path_list)
        number_of_samples = len(path_list)
        number_of_samples_per_block = number_of_samples // self.split_pieces
        dataset_paths = []
        for block in self.dataset_splits:
            if block == (self.split_pieces - 1):
                dataset_paths += path_list[
                                 block * number_of_samples_per_block:]
            else:
                dataset_paths += path_list[
                                 block * number_of_samples_per_block:(block + 1) * number_of_samples_per_block]
        return dataset_paths

    def load_times_and_fluxes_from_path(self, path: Path) -> (np.ndarray, np.ndarray):
        """
        Loads the times and fluxes from a given light curve path.

        :param path: The path to the light curve file.
        :return: The times and the fluxes of the light curve.
        """
        light_curve_dataframe = pd.read_feather(path)
        times = light_curve_dataframe['HJD'].values
        fluxes = light_curve_dataframe['flux'].values
        return times, fluxes


class MicrolensingSyntheticPSPLSignalCollection(LightCurveCollection):
    def __init__(self):
        super().__init__()
        self.label = 1

    def get_paths(self):
        """
        Gets the paths for the PSPL microlensing signals.

        :return: An iterable of the light curve paths.
        """
        paths = Path('/local/data/fugu3/sishitan/muLAn_project/PSPL').glob('*.feather')
        return paths

    def load_times_and_magnifications_from_path(self, path: Path) -> (np.ndarray, np.ndarray):
        """
        Loads the times and magnifications from a given path as an injectable signal.

        :param path: The path to the light curve/signal file.
        :return: The times and the magnifications of the light curve/signal.
        """
        signal_dataframe = pd.read_feather(path)
        times = signal_dataframe['Time'].values
        magnifications = signal_dataframe['Magnification'].values
        return times, magnifications


class MicrolensingSyntheticGeneratedDuringRunningSignalCollection(LightCurveCollection):
    def __init__(self):
        super().__init__()
        self.label = 1

    def get_paths(self) -> Iterable[Path]:
        """
        No need to get paths because this function will generate the signals on the fly.

        :return: empty generator.
        """

        return [Path('')]

    def load_times_and_magnifications_from_path(self, path: Path) -> (np.ndarray, np.ndarray):
        """
        Loads the times and magnifications from a random generated signal.

        :param path: empty path
        :return: The times and the magnifications of the signal.
        """
        random_signal = MagnificationSignal.generate_randomly_based_on_moa_observations()
        return random_signal.timeseries, random_signal.magnification



from dataclasses import dataclass


@dataclass
class TrainHyperparameterConfiguration:
    """
    Hyperparameter configuration settings for a train session.

    :ivar batch_size: The size of the batch for each train process. Each training step will use a number of examples
        equal to this value multiplied by the number of train processes.
    :ivar cycles: The number of train cycles to run.
    """
    learning_rate: float
    optimizer_epsilon: float
    weight_decay: float
    batch_size: int
    cycles: int
    train_steps_per_cycle: int
    validation_steps_per_cycle: int
    norm_based_gradient_clip: float

    @classmethod
    def new(cls,
            learning_rate: float = 1e-4,
            optimizer_epsilon: float = 1e-7,
            weight_decay: float = 0.0001,
            batch_size: int = 100,
            train_steps_per_cycle: int = 100,
            validation_steps_per_cycle: int = 10,
            cycles: int = 5000,
            norm_based_gradient_clip: float = 1.0):
        return cls(learning_rate=learning_rate,
                   optimizer_epsilon=optimizer_epsilon,
                   weight_decay=weight_decay,
                   batch_size=batch_size,
                   cycles=cycles,
                   train_steps_per_cycle=train_steps_per_cycle,
                   validation_steps_per_cycle=validation_steps_per_cycle,
                   norm_based_gradient_clip=norm_based_gradient_clip)



"""
Code for a light curve collection of the TESS FFI data, as produced by Brian Powell.
"""
import numpy as np
from pathlib import Path
from typing import Union, List
from peewee import Select

from ramjet.data_interface.metadatabase import MetadatabaseModel
from ramjet.data_interface.tess_ffi_light_curve_metadata_manager import TessFfiLightCurveMetadataManager, \
    TessFfiLightCurveMetadata
from ramjet.photometric_database.sql_metadata_light_curve_collection import SqlMetadataLightCurveCollection
from ramjet.photometric_database.tess_ffi_light_curve import TessFfiLightCurve


class TessFfiLightCurveCollection(SqlMetadataLightCurveCollection):
    """
    A light curve collection of the TESS two minute cadence data.
    """
    tess_ffi_light_curve_metadata_manger = TessFfiLightCurveMetadataManager()

    def __init__(self, dataset_splits: Union[List[int], None] = None,
                 magnitude_range: (Union[float, None], Union[float, None]) = (None, None)):
        super().__init__()
        self.data_directory: Path = Path('data/tess_ffi_light_curves')
        self.label = 0
        self.dataset_splits: Union[List[int], None] = dataset_splits
        self.magnitude_range: (Union[float, None], Union[float, None]) = magnitude_range

    def get_sql_query(self) -> Select:
        """
        Gets the SQL query for the database models for the light curve collection.

        :return: The SQL query.
        """
        query = TessFfiLightCurveMetadata().select()
        if self.magnitude_range[0] is not None and self.magnitude_range[1] is not None:
            query = query.where(TessFfiLightCurveMetadata.magnitude.between(*self.magnitude_range))
        elif self.magnitude_range[0] is not None:
            query = query.where(TessFfiLightCurveMetadata.magnitude > self.magnitude_range[0])
        elif self.magnitude_range[1] is not None:
            query = query.where(TessFfiLightCurveMetadata.magnitude < self.magnitude_range[1])
        if self.dataset_splits is not None:
            query = query.where(TessFfiLightCurveMetadata.dataset_split.in_(self.dataset_splits))
        return query

    def get_path_from_model(self, model: MetadatabaseModel) -> Path:
        """
        Gets the light curve path from the SQL database model.

        :return: The path to the light curve.
        """
        return Path(self.tess_ffi_light_curve_metadata_manger.
                    light_curve_root_directory_path.joinpath(model.path))

    def load_times_and_fluxes_from_path(self, path: Path) -> (np.ndarray, np.ndarray):
        """
        Loads the times and fluxes from a given light curve path.

        :param path: The path to the light curve file.
        :return: The times and the fluxes of the light curve.
        """
        fluxes, times = TessFfiLightCurve.load_fluxes_and_times_from_pickle_file(path)
        return times, fluxes

    def load_times_and_magnifications_from_path(self, path: Path) -> (np.ndarray, np.ndarray):
        """
        Loads the times and magnifications from a given path as an injectable signal.

        :param path: The path to the light curve/signal file.
        :return: The times and the magnifications of the light curve/signal.
        """
        fluxes, times = TessFfiLightCurve.load_fluxes_and_times_from_pickle_file(path)
        magnifications, times = self.generate_synthetic_signal_from_real_data(fluxes, times)
        return times, magnifications



from ramjet.data_interface.tess_data_interface import TessFluxType
from ramjet.photometric_database.derived.self_lensing_binary_synthetic_signals_light_curve_collection import \
    SelfLensingBinarySyntheticSignalsLightCurveCollection, ReversedSelfLensingBinarySyntheticSignalsLightCurveCollection
from ramjet.photometric_database.derived.tess_two_minute_cadence_light_curve_collection import TessTwoMinuteCadenceLightCurveCollection
from ramjet.photometric_database.standard_and_injected_light_curve_database import StandardAndInjectedLightCurveDatabase


class SelfLensingBinaryDatabase(StandardAndInjectedLightCurveDatabase):
    """
    A database to train a network to find self lensing binaries in TESS two minute cadence data.
    """
    def __init__(self):
        super().__init__()
        self.training_standard_light_curve_collections = [TessTwoMinuteCadenceLightCurveCollection(
            flux_type=TessFluxType.SAP)]
        self.training_injectee_light_curve_collection = TessTwoMinuteCadenceLightCurveCollection(
            flux_type=TessFluxType.SAP)
        self.training_injectable_light_curve_collections = [
            SelfLensingBinarySyntheticSignalsLightCurveCollection(),
            ReversedSelfLensingBinarySyntheticSignalsLightCurveCollection()
        ]
        self.validation_standard_light_curve_collections = self.training_standard_light_curve_collections
        self.validation_injectee_light_curve_collection = self.training_injectee_light_curve_collection
        self.validation_injectable_light_curve_collections = self.training_injectable_light_curve_collections
        self.inference_light_curve_collections = [
            TessTwoMinuteCadenceLightCurveCollection(flux_type=TessFluxType.SAP)]


from pathlib import Path

import pandas as pd
import uvloop
from bokeh import plotting
from bokeh.application import Application
from bokeh.application.handlers import DirectoryHandler
from bokeh.server.server import Server
import argparse

plotting.output_notebook.__doc__ = ''


def run_viewer(light_curve_path: Path, port: int = 5007):
    uvloop.install()
    print(f'Opening viewer on http://localhost:{port}/')
    handler = DirectoryHandler(filename=str(Path(__file__).parent), argv=[light_curve_path])
    application = Application(handler)
    server = Server(application, port=port)
    server.start()
    server.io_loop.add_callback(server.show, "/")
    server.io_loop.start()


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('light_curve_path')
    parser.add_argument('--port', nargs='?', const=5007, type=int)
    args = parser.parse_args()
    light_curve_path = Path(args.light_curve_path)
    if not light_curve_path.exists():
        print(f'File {light_curve_path} not found.')
        raise SystemExit(1)
    run_viewer(light_curve_path, port=args.port)


if __name__ == '__main__':
    main()



from ramjet.photometric_database.derived.tess_two_minute_cadence_light_curve_collection import \
    TessTwoMinuteCadenceTargetDatasetSplitLightCurveCollection
from ramjet.photometric_database.derived.tess_two_minute_cadence_transit_light_curve_collections import \
    TessTwoMinuteCadenceNonTransitLightCurveCollection, \
    TessTwoMinuteCadenceConfirmedTransitLightCurveCollection
from ramjet.photometric_database.standard_and_injected_light_curve_database import \
    StandardAndInjectedLightCurveDatabase
from ramjet.photometric_database.light_curve_dataset_manipulations import OutOfBoundsInjectionHandlingMethod


class TessTwoMinuteCadenceStandardTransitDatabase(StandardAndInjectedLightCurveDatabase):
    """
    A database using standard positive and negative transit light curves.
    """
    def __init__(self):
        super().__init__()
        self.training_standard_light_curve_collections = [
            TessTwoMinuteCadenceConfirmedTransitLightCurveCollection(dataset_splits=list(range(8))),
            TessTwoMinuteCadenceNonTransitLightCurveCollection(dataset_splits=list(range(8)))
        ]
        self.validation_standard_light_curve_collections = [
            TessTwoMinuteCadenceConfirmedTransitLightCurveCollection(dataset_splits=[8]),
            TessTwoMinuteCadenceNonTransitLightCurveCollection(dataset_splits=[8])
        ]
        self.inference_light_curve_collections = [
            TessTwoMinuteCadenceTargetDatasetSplitLightCurveCollection(dataset_splits=[9])]


class TessTwoMinuteCadenceInjectedTransitDatabase(StandardAndInjectedLightCurveDatabase):
    """
    A database using positive and negative transit light curves injected into negative light curves.
    """
    def __init__(self):
        super().__init__()
        self.out_of_bounds_injection_handling = OutOfBoundsInjectionHandlingMethod.REPEAT_SIGNAL
        self.training_injectee_light_curve_collection = TessTwoMinuteCadenceNonTransitLightCurveCollection(
            dataset_splits=list(range(8)))
        self.training_injectable_light_curve_collections = [
            TessTwoMinuteCadenceConfirmedTransitLightCurveCollection(dataset_splits=list(range(8))),
            TessTwoMinuteCadenceNonTransitLightCurveCollection(dataset_splits=list(range(8)))
        ]
        self.validation_standard_light_curve_collections = [
            TessTwoMinuteCadenceConfirmedTransitLightCurveCollection(dataset_splits=[8]),
            TessTwoMinuteCadenceNonTransitLightCurveCollection(dataset_splits=[8])
        ]
        self.inference_light_curve_collections = [
            TessTwoMinuteCadenceTargetDatasetSplitLightCurveCollection(dataset_splits=[9])]


class TessTwoMinuteCadenceStandardAndInjectedTransitDatabase(StandardAndInjectedLightCurveDatabase):
    """
    A database using standard positive and negative transit light curves and positives injected into negatives.
    """
    def __init__(self):
        super().__init__()
        self.out_of_bounds_injection_handling = OutOfBoundsInjectionHandlingMethod.REPEAT_SIGNAL
        self.training_standard_light_curve_collections = [
            TessTwoMinuteCadenceConfirmedTransitLightCurveCollection(dataset_splits=list(range(8))),
            TessTwoMinuteCadenceNonTransitLightCurveCollection(dataset_splits=list(range(8)))
        ]
        self.training_injectee_light_curve_collection = TessTwoMinuteCadenceNonTransitLightCurveCollection(
            dataset_splits=list(range(8)))
        self.training_injectable_light_curve_collections = [
            TessTwoMinuteCadenceConfirmedTransitLightCurveCollection(dataset_splits=list(range(8))),
            TessTwoMinuteCadenceNonTransitLightCurveCollection(dataset_splits=list(range(8)))
        ]
        self.validation_standard_light_curve_collections = [
            TessTwoMinuteCadenceConfirmedTransitLightCurveCollection(dataset_splits=[8]),
            TessTwoMinuteCadenceNonTransitLightCurveCollection(dataset_splits=[8])
        ]
        self.inference_light_curve_collections = [
            TessTwoMinuteCadenceTargetDatasetSplitLightCurveCollection(dataset_splits=[9])]



import math

import torch
from torch import Tensor
from torch.nn import Module, Conv1d, LeakyReLU, Sigmoid, Dropout, Dropout1d, MaxPool1d, BatchNorm1d


class Hadryss(Module):
    def __init__(self):
        super().__init__()
        self.sigmoid = Sigmoid()
        self.block0 = LightCurveNetworkBlock(input_channels=1, output_channels=8, kernel_size=3, pooling_size=2)
        self.block1 = LightCurveNetworkBlock(input_channels=8, output_channels=8, kernel_size=3, pooling_size=2)
        self.block2 = LightCurveNetworkBlock(input_channels=8, output_channels=16, kernel_size=3, pooling_size=2, batch_normalization=True, dropout_rate=0.1)
        self.block3 = LightCurveNetworkBlock(input_channels=16, output_channels=32, kernel_size=3, pooling_size=2, batch_normalization=True, dropout_rate=0.1)
        self.block4 = LightCurveNetworkBlock(input_channels=32, output_channels=64, kernel_size=3, pooling_size=2, batch_normalization=True, dropout_rate=0.1)
        self.block5 = LightCurveNetworkBlock(input_channels=64, output_channels=128, kernel_size=3, pooling_size=2, batch_normalization=True, dropout_rate=0.1)
        self.block6 = LightCurveNetworkBlock(input_channels=128, output_channels=128, kernel_size=3, pooling_size=2, batch_normalization=True, dropout_rate=0.1)
        self.block7 = LightCurveNetworkBlock(input_channels=128, output_channels=128, kernel_size=3, pooling_size=2, batch_normalization=True, dropout_rate=0.1)
        self.block8 = LightCurveNetworkBlock(input_channels=128, output_channels=20, kernel_size=3, pooling_size=1, dropout_rate=0.1,
                                             spatial=False, length=7)
        self.block9 = LightCurveNetworkBlock(input_channels=20, output_channels=20, kernel_size=5, pooling_size=1, dropout_rate=0.1)
        self.block10 = LightCurveNetworkBlock(input_channels=20, output_channels=20, kernel_size=1, pooling_size=1)
        self.prediction_layer = Conv1d(in_channels=20, out_channels=1, kernel_size=1)

    def forward(self, x: Tensor) -> Tensor:
        x = x.reshape([-1, 1, 2500])
        x = self.block0(x)
        x = self.block1(x)
        x = self.block2(x)
        x = self.block3(x)
        x = self.block4(x)
        x = self.block5(x)
        x = self.block6(x)
        x = self.block7(x)
        x = self.block8(x)
        x = self.block9(x)
        x = self.block10(x)
        x = self.prediction_layer(x)
        x = self.sigmoid(x)
        x = torch.reshape(x, (-1,))
        return x


class LightCurveNetworkBlock(Module):
    def __init__(self, input_channels: int, output_channels: int, kernel_size: int, pooling_size: int,
                 dropout_rate: float = 0.0, batch_normalization: bool = False, spatial: bool = True,
                 length: int | None = None):
        super().__init__()
        self.leaky_relu = LeakyReLU()
        self.convolution = Conv1d(in_channels=input_channels, out_channels=output_channels, kernel_size=kernel_size)
        self.spatial: bool = spatial
        self.output_channels: int = output_channels
        if dropout_rate > 0:
            if spatial:
                self.dropout = Dropout1d(dropout_rate)
            else:
                self.dropout = Dropout(dropout_rate)
        else:
            self.dropout = None
        if pooling_size > 1:
            self.max_pooling = MaxPool1d(kernel_size=pooling_size)
        else:
            self.max_pooling = None
        if batch_normalization:
            if spatial:
                self.batch_normalization = BatchNorm1d(num_features=output_channels)
            else:
                assert length is not None
                self.batch_normalization = BatchNorm1d(num_features=output_channels * length)
        else:
            self.batch_normalization = None

    def forward(self, x):
        x = self.convolution(x)
        x = self.leaky_relu(x)
        if self.dropout is not None:
            x = self.dropout(x)
        if self.max_pooling is not None:
            x = self.max_pooling(x)
        if self.batch_normalization is not None:
            if not self.spatial:
                old_shape = x.shape
                x = torch.reshape(x, [-1, torch.prod(torch.tensor(old_shape[1:]))])
            x = self.batch_normalization(x)
            if not self.spatial:
                x = torch.reshape(x, old_shape)
        return x



"""
Code for a light curve collection of Agnieszka Cieplak's synthetic signals.
"""
import re
import tarfile
import urllib.request
import numpy as np
import pandas as pd
from pathlib import Path
from typing import Iterable

from ramjet.photometric_database.light_curve_collection import LightCurveCollection


class SelfLensingBinarySyntheticSignalsLightCurveCollection(LightCurveCollection):
    """
    A light curve collection for Agnieszka Cieplak's synthetic signals.
    """

    def __init__(self):
        super().__init__()
        self.data_directory: Path = Path('data/self_lensing_binary_synthetic_signals')
        self.label = 1

    def download_csv_files(self):
        """
        Downloads Agnieszka Cieplak's synthetic signals in their original CSV form.
        """
        print('Downloading synthetic signal CSV files...')
        tar_file_path = self.data_directory.joinpath('synthetic_signals_csv_files.tar')
        urllib.request.urlretrieve('https://api.onedrive.com/v1.0/shares/s!AjiSFm1N8Bv7ghXushB7JOzABXdv/root/content',
                                   str(tar_file_path))
        with tarfile.open(tar_file_path) as csv_tar_file:
            csv_tar_file.extractall(self.data_directory)
        tar_file_path.unlink()
        csv_uncompressed_directory = self.data_directory.joinpath('LearningSetedgeon_all_sum')
        for path in csv_uncompressed_directory.glob('*'):
            path.rename(self.data_directory.joinpath(path.name))
        csv_uncompressed_directory.rmdir()

    def convert_csv_files_to_project_format(self):
        """
        Converts Agnieszka Cieplak's synthetic signal CSV files to the project format feather files.
        """
        print('Converting synthetic signals to project format...')
        out_paths = self.data_directory.glob('*.out')
        synthetic_signal_csv_paths = [path for path in out_paths if re.match(r'lc_\d+\.out', path.name)]
        for synthetic_signal_csv_path in synthetic_signal_csv_paths:
            synthetic_signal = pd.read_csv(synthetic_signal_csv_path, names=['time__hours', 'magnification'],
                                           delim_whitespace=True, skipinitialspace=True)
            synthetic_signal['time__days'] = synthetic_signal['time__hours'] / 24
            synthetic_signal.drop('time__hours', axis=1, inplace=True)
            synthetic_signal.to_feather(self.data_directory.joinpath(f'{synthetic_signal_csv_path.stem}.feather'))
            synthetic_signal_csv_path.unlink()

    def get_paths(self) -> Iterable[Path]:
        """
        Gets the paths for the light curves in the collection.

        :return: An iterable of the light curve paths.
        """
        return self.data_directory.glob('*.feather')

    def load_times_and_magnifications_from_path(self, path: Path) -> (np.ndarray, np.ndarray):
        """
        Loads the times and magnifications from a given path as an injectable signal.

        :param path: The path to the light curve/signal file.
        :return: The times and the magnifications of the light curve/signal.
        """
        synthetic_signal_data_frame = pd.read_feather(path)
        times = synthetic_signal_data_frame['time__days'].values
        magnifications = synthetic_signal_data_frame['magnification'].values
        return times, magnifications


class ReversedSelfLensingBinarySyntheticSignalsLightCurveCollection(
    SelfLensingBinarySyntheticSignalsLightCurveCollection):
    """
    A light curve collection for a time reversed version Agnieszka Cieplak's synthetic signals. Useful as a negative
    training dataset, as the lensing signals will only occur in a specific order relative to the other signal
    components.
    """

    def __init__(self):
        super().__init__()
        self.data_directory: Path = Path('data/self_lensing_binary_synthetic_signals')
        self.label = 0

    @staticmethod
    def reverse_signal(times: np.ndarray, magnifications: np.ndarray) -> (np.ndarray, np.ndarray):
        """
        Reverses the signal (i.e., the original first magnitude is in the last position and vice versa).

        :param times: The times of the original signal.
        :param magnifications: The magnifications of the original signal.
        :return: The reversed signal times and magnitudes.
        """
        return times, magnifications[::-1]

    def load_times_and_magnifications_from_path(self, path: Path) -> (np.ndarray, np.ndarray):
        """
        Loads the times and magnifications from a given path as an injectable signal.

        :param path: The path to the light curve/signal file.
        :return: The times and the magnifications of the light curve/signal.
        """
        original_times, original_magnitudes = super().load_times_and_magnifications_from_path(path)
        times, magnifications = self.reverse_signal(original_times, original_magnitudes)
        return times, magnifications


if __name__ == '__main__':
    light_curve_collection = SelfLensingBinarySyntheticSignalsLightCurveCollection()
    light_curve_collection.data_directory.mkdir(parents=True, exist_ok=True)
    light_curve_collection.download_csv_files()
    light_curve_collection.convert_csv_files_to_project_format()
    print('Self lensing binary synthetic signal light curve collection ready.')



from __future__ import annotations

from typing import Optional

import numpy as np
from astropy import units
from bokeh.document import Document
from bokeh.events import Tap
from bokeh.palettes import Turbo256
from lightkurve.periodogram import LombScarglePeriodogram, Periodogram

try:
    from enum import StrEnum
except ImportError:
    from backports.strenum import StrEnum

from bokeh.models import Spinner, ColumnDataSource, LinearColorMapper, TapTool, Span, Range1d, LinearAxis, Div
from bokeh.plotting import figure as Figure

from ramjet.photometric_database.light_curve import LightCurve


class FoldedLightCurveColumnName(StrEnum):
    TIME = 'time'
    FOLDED_TIME = 'folded_time'
    FLUX = 'flux'


class PeriodogramColumnName(StrEnum):
    PERIOD = 'period'
    POWER = 'power'


class Viewer:
    def __init__(self, bokeh_document: Document, light_curve: LightCurve, title: Optional[str] = None):
        self.bokeh_document: Document = bokeh_document
        tool_tips = [
            ("Time", f"@{FoldedLightCurveColumnName.TIME}{{0.0000000}}"),
            ("Folded time", f"@{FoldedLightCurveColumnName.FOLDED_TIME}{{0.0000000}}"),
            ("Flux", f"@{FoldedLightCurveColumnName.FLUX}{{0.0000000}}"),
        ]
        self.folded_light_curve_figure: Figure = Figure(tooltips=tool_tips)
        self.folded_light_curve_figure.sizing_mode = 'stretch_width'
        self.unfolded_light_curve_figure: Figure = Figure(tooltips=tool_tips)
        self.unfolded_light_curve_figure.sizing_mode = 'stretch_width'
        self.light_curve: LightCurve = light_curve
        flux_median = np.nanmedian(self.light_curve.fluxes)
        fluxes = self.light_curve.fluxes
        relative_fluxes = fluxes / flux_median
        minimum_time = np.nanmin(self.light_curve.times)
        maximum_time = np.nanmax(self.light_curve.times)
        time_differences = np.diff(self.light_curve.times)
        minimum_time_step = np.nanmin(time_differences)
        median_time_step = np.nanmedian(time_differences)
        period_upper_limit = maximum_time - minimum_time
        period_lower_limit = minimum_time_step / 2.1
        self.fold_period_spinner: Spinner = Spinner(value=period_upper_limit, low=period_lower_limit,
                                                    high=period_upper_limit, step=median_time_step / 1000)
        self.fold_period_spinner.on_change('value', self.update_fold)
        mapper = LinearColorMapper(palette=Turbo256, low=minimum_time, high=maximum_time)
        color = {'field': FoldedLightCurveColumnName.TIME, 'transform': mapper}

        self.unfolded_light_curve_column_data_source: ColumnDataSource = ColumnDataSource(data={
            FoldedLightCurveColumnName.TIME: self.light_curve.times,
            FoldedLightCurveColumnName.FOLDED_TIME: self.light_curve.times,
            FoldedLightCurveColumnName.FLUX: relative_fluxes,
        })
        self.unfolded_light_curve_figure.circle(source=self.unfolded_light_curve_column_data_source,
                                              x=FoldedLightCurveColumnName.FOLDED_TIME,
                                              y=FoldedLightCurveColumnName.FLUX, line_color=color, line_alpha=0.8,
                                              fill_color=color, fill_alpha=0.2)

        self.folded_light_curve_column_data_source: ColumnDataSource = ColumnDataSource(data={
            FoldedLightCurveColumnName.TIME: self.light_curve.times,
            FoldedLightCurveColumnName.FOLDED_TIME: self.light_curve.times,
            FoldedLightCurveColumnName.FLUX: relative_fluxes,
        })
        self.folded_light_curve_figure.circle(source=self.folded_light_curve_column_data_source,
                                              x=FoldedLightCurveColumnName.FOLDED_TIME,
                                              y=FoldedLightCurveColumnName.FLUX, line_color=color, line_alpha=0.8,
                                              fill_color=color, fill_alpha=0.2)

        fluxes_minimum = fluxes.min()
        fluxes_maximum = fluxes.max()
        fluxes_margin = (fluxes_maximum - fluxes_minimum) * 0.05
        fluxes_range = Range1d(fluxes_minimum - fluxes_margin, fluxes_maximum + fluxes_margin)

        relative_fluxes_minimum = relative_fluxes.min()
        relative_fluxes_maximum = relative_fluxes.max()
        relative_fluxes_margin = (relative_fluxes_maximum - relative_fluxes_minimum) * 0.05
        relative_fluxes_range = Range1d(relative_fluxes_minimum - relative_fluxes_margin,
                                             relative_fluxes_maximum + relative_fluxes_margin)

        self.folded_light_curve_figure.y_range = relative_fluxes_range
        self.folded_light_curve_figure.extra_y_ranges = {'unnormalized_range': fluxes_range}
        self.folded_light_curve_figure.add_layout(LinearAxis(y_range_name='unnormalized_range'), 'right')

        self.periodogram_figure: Figure = Figure(active_drag='box_zoom')
        self.current_fold_period_span = Span(location=self.fold_period_spinner.value, dimension='height',
                                        line_color='firebrick')
        self.periodogram_figure.add_layout(self.current_fold_period_span)
        self.periodogram_figure.add_tools(TapTool())
        self.periodogram_figure.on_event(Tap, self.periodogram_tap_callback)
        lightkurve_light_curve = self.light_curve.to_lightkurve()
        inlier_lightkurve_light_curve = lightkurve_light_curve.remove_outliers(sigma=3)
        periodogram: Periodogram = LombScarglePeriodogram.from_lightcurve(inlier_lightkurve_light_curve,
                                                                          oversample_factor=5,
                                                                          minimum_period=period_lower_limit,
                                                                          maximum_period=period_upper_limit)
        periods__days = periodogram.period.to(units.day).value
        powers = periodogram.power.value

        self.periodogram_column_data_source: ColumnDataSource = ColumnDataSource(data={
            PeriodogramColumnName.PERIOD: periods__days,
            PeriodogramColumnName.POWER: powers,
        })
        self.periodogram_figure.line(source=self.periodogram_column_data_source,
                                     x=PeriodogramColumnName.PERIOD,
                                     y=PeriodogramColumnName.POWER)
        self.periodogram_figure.sizing_mode = 'stretch_width'

        if title is not None:
            title_div = Div(text=f'<h1>{title}</h1>')
            self.bokeh_document.add_root(title_div)
        self.bokeh_document.add_root(self.unfolded_light_curve_figure)
        self.bokeh_document.add_root(self.folded_light_curve_figure)
        self.bokeh_document.add_root(self.fold_period_spinner)
        self.bokeh_document.add_root(self.periodogram_figure)

        # TODO: Remove spike.
        # Spike to make the default fold and zoom be useful for the short period application.
        longest_period_index_near_max_power = np.argwhere(powers > 0.9 * periodogram.max_power)[0, -1]
        while powers[longest_period_index_near_max_power + 1] > powers[longest_period_index_near_max_power]:
            longest_period_index_near_max_power += 1
        longest_period_near_max_power = periods__days[longest_period_index_near_max_power]
        self.fold_period_spinner.value = longest_period_near_max_power
        self.periodogram_figure.x_range.start = period_lower_limit
        self.periodogram_figure.x_range.end = longest_period_near_max_power * 2
        self.update_view_with_new_fold()

    def update_fold(self, attr, old, new):
        self.calculate_new_fold()
        self.update_view_with_new_fold()

    def calculate_new_fold(self):
        self.light_curve.fold(self.fold_period_spinner.value, epoch=0)

    def update_view_with_new_fold(self):
        self.current_fold_period_span.location = self.fold_period_spinner.value
        self.folded_light_curve_column_data_source.data[
            FoldedLightCurveColumnName.FOLDED_TIME] = self.light_curve.folded_times

    def periodogram_tap_callback(self, event):
        self.fold_period_spinner.value = event.x
        self.update_view_with_new_fold()




"""
Code to represent a TESS target.
"""
from __future__ import annotations

import io
import math
import numpy as np
import pandas as pd
from typing import Union

import requests
from astroquery.gaia import Gaia

from ramjet.data_interface.tess_data_interface import get_tess_input_catalog_row


class TessTarget:
    """
    A class to represent an TESS target. Usually a star or star system.
    """
    def __init__(self):
        self.tic_id: Union[int, None] = None
        self.radius: Union[float, None] = None
        self.mass: Union[float, None] = None
        self.magnitude: Union[float, None] = None
        self.contamination_ratio: Union[float, None] = None

    @classmethod
    def from_tic_id(cls, tic_id: int) -> TessTarget:
        """
        Creates a target from a TIC ID.

        :param tic_id: The TIC ID to create the target from.
        :return: The target.
        """
        target = TessTarget()
        target.tic_id = tic_id
        tic_row = get_tess_input_catalog_row(target.tic_id)
        target.radius = tic_row['rad']
        if np.isnan(target.radius):
            gaia_source_id_string = tic_row['GAIA']
            if pd.notna(gaia_source_id_string):
                target.radius = target.get_radius_from_gaia(int(gaia_source_id_string))
        target.mass = tic_row['mass']
        # noinspection SpellCheckingInspection
        target.magnitude = tic_row['Tmag']
        # noinspection SpellCheckingInspection
        target.contamination_ratio = tic_row['contratio']
        return target

    @staticmethod
    def get_radius_from_gaia(gaia_source_id: int) -> float:
        """
        Retrieves the radius of a body from Gaia.

        :param gaia_source_id: The Gaia source ID for the target to retrieve the radius of.
        :return: The radius.
        """
        # noinspection SqlResolve
        gaia_job = Gaia.launch_job(f'select * from gaiadr2.gaia_source where source_id={gaia_source_id}')
        query_results_data_frame = gaia_job.get_results().to_pandas()
        radius = query_results_data_frame['radius_val'].iloc[0]
        return radius

    def calculate_transiting_body_radius(self, transit_depth: float, allow_unknown_contamination_ratio: bool = False
                                         ) -> float:
        """
        Calculates the radius of a transiting body based on the target parameters and the transit depth.

        :param transit_depth: The depth of the transit signal.
        :param allow_unknown_contamination_ratio: Whether to calculate even without a known contamination ratio.
        :return: The calculated radius of the transiting body.
        """
        contamination_ratio = self.contamination_ratio
        if pd.isna(contamination_ratio):
            if allow_unknown_contamination_ratio:
                contamination_ratio = 0
            else:
                raise ValueError(f'Contamination ratio {contamination_ratio} cannot be used to calculate the radius.')
        return self.radius * math.sqrt(transit_depth * (1 + contamination_ratio))

    def retrieve_nearby_tic_targets(self) -> pd.DataFrame:
        """
        Retrieves the data frame of nearby targets from ExoFOP.

        :return: The data frame of nearby targets.
        """
        csv_url = f'https://exofop.ipac.caltech.edu/tess/download_nearbytarget.php?id={self.tic_id}&output=csv'
        csv_string = requests.get(csv_url).content.decode('utf-8')
        if 'Distance Err' not in csv_string:  # Correct ExoFOP bug where distance error column header is missing.
            csv_string = csv_string.replace('Distance(pc)', 'Distance (pc),Distance Err (pc)')
        data_frame = pd.read_csv(io.StringIO(csv_string), index_col=False)
        data_frame = data_frame[data_frame['TIC ID'] != self.tic_id]  # Remove row of current target.
        return data_frame



from typing import Tuple

import torch
import numpy.typing as npt
import numpy as np
from torch import Tensor

from qusi.light_curve_observation import LightCurveObservation


def from_light_curve_observation_to_fluxes_array_and_label_array(light_curve_observation: LightCurveObservation
                                               ) -> (npt.NDArray[np.float32], npt.NDArray[np.float32]):
    fluxes = light_curve_observation.light_curve.fluxes
    label = light_curve_observation.label
    return fluxes, np.array(label, dtype=np.float32)

def pair_array_to_tensor(arrays: Tuple[npt.NDArray[np.float32], npt.NDArray[np.float32]]) -> (Tensor, Tensor):
    return torch.tensor(arrays[0]), torch.tensor(arrays[1])



"""
Code for managing the meta data of the two minute cadence TESS light curves.
"""
from pathlib import Path
from typing import List
from peewee import IntegerField, CharField, SchemaManager

from ramjet.data_interface.metadatabase import MetadatabaseModel, metadatabase, metadatabase_uuid, \
    convert_class_to_table_name, dataset_split_from_uuid
from ramjet.data_interface.tess_data_interface import get_tic_id_and_sector_from_file_path


class TessTwoMinuteCadenceLightCurveMetadata(MetadatabaseModel):
    """
    A model for the TESS two minute cadence light curve metadatabase table.
    """
    tic_id = IntegerField(index=True)
    sector = IntegerField(index=True)
    path = CharField(unique=True)
    dataset_split = IntegerField()

    class Meta:
        """Schema meta data for the model."""
        indexes = (
            (('sector', 'tic_id'), True),  # Ensures TIC ID and sector entry is unique.
        )


class TessTwoMinuteCadenceLightCurveMetadataManger:
    """
    A class for managing the metadata of the two-minute cadence TESS light curves.
    """
    def __init__(self):
        self.light_curve_root_directory_path = Path('data/tess_two_minute_cadence_light_curves')

    def insert_multiple_rows_from_paths_into_database(self, light_curve_paths: List[Path]):
        """
        Inserts sets of light curve paths into the table.

        :param light_curve_paths: The list of paths to insert.
        """
        row_dictionary_list = []
        table_name = convert_class_to_table_name(TessTwoMinuteCadenceLightCurveMetadata)
        for light_curve_path in light_curve_paths:
            tic_id, sector = get_tic_id_and_sector_from_file_path(light_curve_path)
            uuid_name = f'{table_name} TIC {tic_id} sector {sector}'
            uuid = metadatabase_uuid(uuid_name)
            dataset_split = dataset_split_from_uuid(uuid)
            row_dictionary_list.append({TessTwoMinuteCadenceLightCurveMetadata.path.name: str(light_curve_path),
                                        TessTwoMinuteCadenceLightCurveMetadata.tic_id.name: tic_id,
                                        TessTwoMinuteCadenceLightCurveMetadata.sector.name: sector,
                                        TessTwoMinuteCadenceLightCurveMetadata.dataset_split.name: dataset_split})
        with metadatabase.atomic():
            TessTwoMinuteCadenceLightCurveMetadata.insert_many(row_dictionary_list).execute()

    def populate_sql_database(self):
        """
        Populates the SQL database based on the light curve files.
        """
        print('Populating the TESS two minute cadence light curve meta data table...')
        path_glob = self.light_curve_root_directory_path.glob('**/*.fits')
        row_count = 0
        batch_paths = []
        with metadatabase.atomic():
            for index, path in enumerate(path_glob):
                batch_paths.append(path.relative_to(self.light_curve_root_directory_path))
                row_count += 1
                if index % 1000 == 0 and index != 0:
                    self.insert_multiple_rows_from_paths_into_database(batch_paths)
                    batch_paths = []
                    print(f'{index} rows inserted...', end='\r')
            if len(batch_paths) > 0:
                self.insert_multiple_rows_from_paths_into_database(batch_paths)
        print(f'TESS two minute cadence light curve meta data table populated. {row_count} rows added.')

    def build_table(self):
        """
        Builds the SQL table.
        """
        TessTwoMinuteCadenceLightCurveMetadata.drop_table()
        TessTwoMinuteCadenceLightCurveMetadata.create_table()
        SchemaManager(TessTwoMinuteCadenceLightCurveMetadata).drop_indexes()  # To allow for fast insert.
        self.populate_sql_database()
        print('Building indexes...')
        SchemaManager(TessTwoMinuteCadenceLightCurveMetadata).create_indexes()  # Since we dropped them before.


if __name__ == '__main__':
    manager = TessTwoMinuteCadenceLightCurveMetadataManger()
    manager.build_table()



"""
Code for displaying a light curve figure.
"""
import numpy as np
import pandas as pd
from typing import Union, List

from bokeh.colors import Color
from bokeh.models import ColumnDataSource, GlyphRenderer, CustomJS
from bokeh.plotting import figure as Figure

from ramjet.analysis.color_palette import light_curve_colors
from ramjet.analysis.light_curve_visualizer import calculate_inlier_range
from ramjet.analysis.convert_column_name_to_display_name import convert_column_name_to_display_name
from ramjet.photometric_database.light_curve import LightCurve


class LightCurveDisplay:
    """
    A class for displaying a light curve figure.
    """
    def __init__(self):
        self.figure: Union[Figure, None] = None
        self.data_source: Union[ColumnDataSource, None] = None
        self.time_column_name: Union[str, None] = None
        self.flux_column_names: List[str] = []
        self.inlier_range_data_source: Union[ColumnDataSource, None] = None
        self.inlier_range_glyph_renderer: Union[GlyphRenderer, None] = None
        self.exclude_outliers_from_zoom: bool = False

    @classmethod
    def for_columns(cls, time_column_name: str, flux_column_names: List[str], flux_axis_label: str = 'Flux'):
        """
        Creates a light curve display object with the specified columns prepared for display.

        :param time_column_name: The name of the time column in the data.
        :param flux_column_names: The names of the flux columns in the data.
        :param flux_axis_label: The name to display on the flux axis.
        :return: The light curve display.
        """
        display = cls()
        display.time_column_name = time_column_name
        display.flux_column_names = flux_column_names
        time_axis_label = convert_column_name_to_display_name(time_column_name)
        display.initialize_figure(time_axis_label=time_axis_label, flux_axis_label=flux_axis_label)
        display.initialize_data_source(column_names=[time_column_name] + flux_column_names)
        for flux_column_name, color in zip(display.flux_column_names, light_curve_colors):
            display.add_flux_data_source_line_to_figure(time_column_name=time_column_name, flux_column_name=flux_column_name,
                                                        color=color)
        return display

    def initialize_figure(self, time_axis_label: str, flux_axis_label: str):
        """
        Initializes the figure.

        :param time_axis_label: The time axis label.
        :param flux_axis_label: The flux axis label.
        """
        self.figure = Figure(x_axis_label=time_axis_label, y_axis_label=flux_axis_label,
                             active_drag='box_zoom', active_scroll='wheel_zoom')
        self.figure.sizing_mode = 'stretch_width'

    def initialize_data_source(self, column_names: List[str]):
        """
        Creates a data source with the passed column names.

        :param column_names: The column names to include in the data source.
        """
        self.data_source = ColumnDataSource(data=pd.DataFrame({column_name: [] for column_name in column_names}))
        js_reset = CustomJS(args=dict(figure=self.figure), code='figure.reset.emit()')
        self.data_source.js_on_change('data', js_reset)

    def add_flux_data_source_line_to_figure(self, time_column_name: str, flux_column_name: str, color: Color):
        """
        Add a flux data source time series line to the figure.

        :param time_column_name: The name to use for the time column.
        :param flux_column_name: The name to use for the flux column.
        :param color: The color to use for the line.
        """
        legend_label = convert_column_name_to_display_name(flux_column_name)
        self.figure.line(x=time_column_name, y=flux_column_name, source=self.data_source, line_color=color,
                         line_alpha=0.1)
        self.figure.circle(x=time_column_name, y=flux_column_name, source=self.data_source, legend_label=legend_label,
                           line_color=color, line_alpha=0.4, fill_color=color, fill_alpha=0.1)

    async def update_from_light_curve(self, light_curve: LightCurve):
        """
        Update the data for the display based on a light curve.

        :param light_curve: The light curve to display.
        """
        self.data_source.data = light_curve.data_frame
        if self.exclude_outliers_from_zoom:
            y_range_minimum, y_range_maximum = await calculate_inlier_range(light_curve.fluxes)
            await self.set_view_ranges(x_range=(np.nanmin(light_curve.times), np.nanmax(light_curve.times)),
                                       y_range=(y_range_minimum, y_range_maximum))

    async def set_view_ranges(self, x_range: (float, float), y_range: (float, float)):
        """
        Sets the view range to a specified range.

        :param x_range: The start and end of the x range.
        :param y_range: The start and end of the y range.
        """
        if self.inlier_range_glyph_renderer is None:
            await self.enable_auto_range_selection()
        data_frame = pd.DataFrame({'x': x_range, 'y': y_range})
        self.inlier_range_data_source.data = data_frame

    async def enable_auto_range_selection(self):
        """
        Enables selecting a specific auto range by creating a dummy invisible glyph and auto ranging to that.
        """
        initial_data_frame = pd.DataFrame({'x': [], 'y': []})
        self.inlier_range_data_source = ColumnDataSource(data=initial_data_frame)
        self.inlier_range_glyph_renderer = self.figure.circle(x='x', y='y', source=self.inlier_range_data_source,
                                                              alpha=0)
        self.figure.y_range.renderers = [self.inlier_range_glyph_renderer]



"""
Code for managing the metadata of the TESS targets.
"""
from pathlib import Path
from typing import List
from peewee import IntegerField, SchemaManager

from ramjet.data_interface.metadatabase import MetadatabaseModel, metadatabase, convert_class_to_table_name, \
    metadatabase_uuid, dataset_split_from_uuid
from ramjet.data_interface.tess_data_interface import TessDataInterface


class TessTargetMetadata(MetadatabaseModel):
    """
    A model for the TESS target metadatabase table.
    """
    tic_id = IntegerField(index=True, unique=True)
    dataset_split = IntegerField()


class TessTargetMetadataManger:
    """
    A class for managing the metadata of TESS targets.
    """
    tess_data_interface = TessDataInterface()

    def __init__(self):
        self.light_curve_root_directory_path = Path('data/tess_two_minute_cadence_light_curves')

    def insert_multiple_rows_from_paths_into_database(self, light_curve_paths: List[Path]) -> int:
        """
        Inserts sets targets into the table from light curve paths.

        :param light_curve_paths: The list of paths to insert.
        :return: The number of rows inserted.
        """
        row_dictionary_list = []
        table_name = convert_class_to_table_name(TessTargetMetadata)
        for light_curve_path in light_curve_paths:
            tic_id, _ = get_tic_id_and_sector_from_file_path(light_curve_path)
            uuid_name = f'{table_name} TIC {tic_id}'
            uuid = metadatabase_uuid(uuid_name)
            dataset_split = dataset_split_from_uuid(uuid)
            row_dictionary_list.append({TessTargetMetadata.tic_id.name: tic_id,
                                        TessTargetMetadata.dataset_split.name: dataset_split})
        with metadatabase.atomic():
            number_of_rows_inserted = TessTargetMetadata.insert_many(row_dictionary_list).on_conflict_ignore().execute()
        return number_of_rows_inserted

    def populate_sql_database(self):
        """
        Populates the SQL database based on the light curve files.
        """
        print('Populating the TESS target light curve metadata table...')
        path_glob = self.light_curve_root_directory_path.glob('**/*.fits')
        row_count = 0
        batch_paths = []
        batch_dataset_splits = []
        with metadatabase.atomic():
            for index, path in enumerate(path_glob):
                batch_paths.append(path.relative_to(self.light_curve_root_directory_path))
                batch_dataset_splits.append(index % 10)
                if index % 1000 == 0 and index != 0:
                    row_count += self.insert_multiple_rows_from_paths_into_database(batch_paths)
                    batch_paths = []
                    batch_dataset_splits = []
                    print(f'{row_count} rows inserted...', end='\r')
            if len(batch_paths) > 0:
                row_count += self.insert_multiple_rows_from_paths_into_database(batch_paths)
        print(f'TESS target metadata table populated. {row_count} rows added.')

    def build_table(self):
        """
        Builds the SQL table.
        """
        TessTargetMetadata.drop_table()
        TessTargetMetadata.create_table()
        self.populate_sql_database()


if __name__ == '__main__':
    manager = TessTargetMetadataManger()
    manager.build_table()



"""
Code for a model for the TESS target database table.
"""
from typing import List

from peewee import IntegerField, AutoField

from ramjet.database.base_model import BaseModel


class TessTarget(BaseModel):
    """
    A model for the TESS target database table.
    """
    id = AutoField()
    tic_id = IntegerField(index=True, unique=True)



"""
Code for the the base model of the database to build other models from.
"""
from typing import Type

from peewee import Model, SqliteDatabase

database = SqliteDatabase('data/database.sqlite3',
                          pragmas={'journal_mode': 'wal'},
                          check_same_thread=False)


def convert_class_to_table_name(model_class: Type[Model]):
    """
    Creates the table name based on the model class.

    :param model_class: The class to create the table name for.
    :return: The name of the table.
    """
    model_name = model_class.__name__
    return model_name


class BaseModel(Model):
    """
    A general model for the database tables.
    """
    class Meta:
        """The meta information for the database models."""
        database = database
        table_function = convert_class_to_table_name



from pathlib import Path
from typing import Iterable, Union

import numpy as np

from ramjet.photometric_database.light_curve import LightCurve
from ramjet.photometric_database.light_curve_collection import LightCurveCollection


class ToyLightCurveCollection(LightCurveCollection):
    """
    A collection of simple toy light curves.
    """
    def get_paths(self) -> Iterable[Path]:
        """
        Gets the paths for the light curves in the collection.

        :return: An iterable of the light curve paths.
        """
        return [Path('')]


class ToyFlatLightCurveCollection(ToyLightCurveCollection):
    def __init__(self):
        super().__init__()
        self.label = 0

    def load_times_and_fluxes_from_path(self, path: Path) -> (np.ndarray, np.ndarray):
        light_curve = ToyLightCurve.flat()
        return light_curve.times, light_curve.fluxes


class ToyFlatAtValueLightCurveCollection(ToyLightCurveCollection):
    def __init__(self):
        super().__init__()

    def get_paths(self) -> Iterable[Path]:
        paths = [Path(f'{index}') for index in range(10)]
        return paths

    def load_times_and_fluxes_from_path(self, path: Path) -> (np.ndarray, np.ndarray):
        light_curve = ToyLightCurve.flat(float(path.name))
        return light_curve.times, light_curve.fluxes

    def load_label_from_path(self, path: Path) -> Union[float, np.ndarray]:
        label = float(path.name)
        return label


class ToySineWaveLightCurveCollection(ToyLightCurveCollection):
    def __init__(self):
        super().__init__()
        self.label = 1

    def load_times_and_fluxes_from_path(self, path: Path) -> (np.ndarray, np.ndarray):
        light_curve = ToyLightCurve.sine_wave()
        return light_curve.times, light_curve.fluxes


class ToyLightCurve:
    """
    Simple toy light curves.
    """

    @classmethod
    def flat(cls, value: float = 1) -> LightCurve:
        """
        Creates a flat light curve.
        """
        length = 100
        fluxes = np.full(shape=[length], fill_value=value, dtype=np.float32)
        times = np.arange(length, dtype=np.float32)
        return LightCurve.from_times_and_fluxes(times=times, fluxes=fluxes)

    @classmethod
    def sine_wave(cls, period=50) -> LightCurve:
        """
        Creates a sine wave light curve.
        """
        length = 100
        periods_to_produce = length / period
        fluxes = (np.sin(np.linspace(0, np.pi * periods_to_produce, num=length, endpoint=False)) / 2) + 1
        times = np.arange(length)
        return LightCurve.from_times_and_fluxes(times=times, fluxes=fluxes)



"""
Code for loading trained models.
"""
import re
import datetime
from pathlib import Path
from typing import Union


def get_latest_log_directory(logs_directory: Union[Path, str]) -> Path:
    """
    Gets the most recent log directory in a root logs directory based on the datetime in the log name.

    :param logs_directory: The root logs directory containing the individual log directories.
    :return: The string path to the most recent log directory.
    """
    logs_directory = Path(logs_directory)
    log_directories = [path for path in Path(logs_directory).glob('*') if path.is_dir()]
    latest_log_directory = None
    latest_log_datetime = datetime.datetime.min
    for log_directory in log_directories:
        match = re.search(r'(\d{4}-\d{2}-\d{2}-\d{2}-\d{2}-\d{2})', str(log_directory))
        if match is not None:
            log_datetime = datetime.datetime.strptime(match.group(1), '%Y-%m-%d-%H-%M-%S')
            if log_datetime > latest_log_datetime:
                latest_log_directory = log_directory
                latest_log_datetime = log_datetime
    if latest_log_directory is None:
        raise FileNotFoundError(f'No logs with datetime names found in {logs_directory}')
    else:
        return latest_log_directory



"""
Code to download and prepare the data needed for the GPU experiments.
"""
from pathlib import Path

from ramjet.data_interface.tess_data_interface import TessDataInterface
from ramjet.data_interface.tess_toi_data_interface import TessToiDataInterface
from ramjet.photometric_database.setup.tess_two_minute_cadence_transit_metadata import build_tables


def download():
    """
    Downloads and prepares the data needed for the GPU experiments.
    """
    tess_data_interface = TessDataInterface()
    download_two_minute_cadence_light_curves(Path('data/tess_two_minute_cadence_light_curves'))
    build_tables()


if __name__ == '__main__':
    download()



from dataclasses import dataclass
from typing import Any, Dict


@dataclass
class TrainLoggingConfiguration:
    """
    Configuration settings for the logging of a train session.

    :ivar wandb_project: The wandb project to log to.
    :ivar wandb_entity: The wandb entity to log to.
    :ivar additional_log_dictionary: The dictionary of additional values to log.
    """
    wandb_project: str | None
    wandb_entity: str | None
    additional_log_dictionary: Dict[str, Any]

    @classmethod
    def new(cls,
            wandb_project: str | None = None,
            wandb_entity: str | None = None,
            additional_log_dictionary: Dict[str, Any] | None = None):
        if additional_log_dictionary is None:
            additional_log_dictionary = {}
        return cls(wandb_project=wandb_project,
                   wandb_entity=wandb_entity,
                   additional_log_dictionary=additional_log_dictionary)



"""
Code for a logging agent to wandb.
"""
from __future__ import annotations

import math
import queue
from abc import ABC, abstractmethod
from pathlib import Path

import plotly
import wandb
from typing import Optional, Dict
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from pathos.helpers import mp as multiprocess

from ramjet.photometric_database.light_curve import LightCurve


class ExampleRequest:
    """
    A representation of a request for an example.
    """
    pass


class WandbLoggable(ABC):
    """
    An object which is loggable to wandb.
    """

    @abstractmethod
    def log(self, summary_name: str, epoch: int) -> None:
        """
        Logs self to wandb.

        :param summary_name: The name of the summary to use on wandb.
        :param epoch: The current epoch to log with.
        """
        raise NotImplementedError

    @staticmethod
    def log_figure(summary_name: str, figure: plotly.graph_objects.Figure, epoch: int) -> None:
        """
        Logs a figure to wandb.

        :param summary_name: The name of the summary to use on wandb.
        :param figure: The figure to be logged.
        :param epoch: The current epoch to log with.
        """
        wandb.log({summary_name: wandb.Plotly(figure)}, step=epoch)


class WandbLoggableLightCurve(WandbLoggable):
    """
    A wandb loggable light curve.
    """
    def __init__(self, light_curve_name: str, light_curve: LightCurve):
        super().__init__()
        self.light_curve_name: str = light_curve_name
        self.light_curve: LightCurve = light_curve

    def log(self, summary_name: str, epoch: int) -> None:
        """
        Logs self to wandb.

        :param summary_name: The name of the summary to use on wandb.
        :param epoch: The current epoch to log with.
        """
        figure = go.Figure()
        figure.add_trace(go.Scatter(x=self.light_curve.times,
                                    y=self.light_curve.fluxes,
                                    mode='lines+markers'))
        figure.update_layout(title=self.light_curve_name,
                             margin={'l': 0, 'r': 0, 'b': 0, 't': 0})
        self.log_figure(summary_name, figure, epoch)


class WandbLoggableInjection(WandbLoggable):
    """
    A wandb loggable containing logging data for injecting a signal into a light curve.
    """
    def __init__(self):
        super().__init__()
        self.injectee_name: Optional[str] = None
        self.injectee_light_curve: Optional[LightCurve] = None
        self.injectable_name: Optional[str] = None
        self.injectable_light_curve: Optional[LightCurve] = None
        self.injected_light_curve: Optional[LightCurve] = None
        self.injectable_light_curve: Optional[LightCurve] = None
        self.injected_light_curve: Optional[LightCurve] = None
        self.aligned_injectable_light_curve: Optional[LightCurve] = None
        self.aligned_injectee_light_curve: Optional[LightCurve] = None
        self.aligned_injected_light_curve: Optional[LightCurve] = None

    def log(self, summary_name: str, epoch: int):
        """
        Logs self to wandb.

        :param summary_name: The name of the summary to use on wandb.
        :param epoch: The current epoch to log with.
        """
        figure = make_subplots(rows=3, cols=1)
        figure.add_trace(go.Scatter(x=self.injectee_light_curve.times,
                                    y=self.injectee_light_curve.fluxes,
                                    mode='lines+markers', name="injectee"), row=1, col=1)
        figure.add_trace(go.Scatter(x=self.injectable_light_curve.times,
                                    y=self.injectable_light_curve.fluxes,
                                    mode='lines+markers', name="injectable"), row=2, col=1)
        figure.add_trace(go.Scatter(x=self.injected_light_curve.times,
                                    y=self.injected_light_curve.fluxes,
                                    mode='lines+markers', name="injected"), row=3, col=1)
        figure.update_layout(title_text=f"{self.injectable_name} injected into "
                                        f"{self.injectee_name}",
                             margin={'l': 0, 'r': 0, 'b': 0, 't': 0})
        self.log_figure(summary_name, figure, epoch)
        aligned_figure = make_subplots(rows=2, cols=1, shared_xaxes=True)
        aligned_figure.add_trace(go.Scatter(x=self.aligned_injectee_light_curve.times,
                                            y=self.aligned_injectee_light_curve.fluxes,
                                            mode='lines+markers', name="injectee"), row=1, col=1)
        aligned_figure.add_trace(go.Scatter(x=self.aligned_injectable_light_curve.times,
                                            y=self.aligned_injectable_light_curve.fluxes,
                                            mode='lines+markers', name="injectable"), row=2, col=1)
        aligned_figure.add_trace(go.Scatter(x=self.aligned_injected_light_curve.times,
                                            y=self.aligned_injected_light_curve.fluxes,
                                            mode='lines+markers', name="injected"), row=1, col=1)
        aligned_figure.update_layout(title_text=f"Aligned {self.injectable_name} injected into "
                                                f"{self.injectee_name}",
                                     margin={'l': 0, 'r': 0, 'b': 0, 't': 0})
        aligned_summary_name = summary_name + 'aligned'
        self.log_figure(aligned_summary_name, aligned_figure, epoch)


class WandbLogger:
    """
    A class to log to wandb.
    """
    loggable_types = [LightCurve]

    def __init__(self):
        manager = multiprocess.Manager()
        self.lock = manager.Lock()
        self.request_queues: Dict[str, multiprocess.Queue] = {}
        self.example_queues: Dict[str, multiprocess.Queue] = {}

    @classmethod
    def new(cls, entity: Optional[str] = None, project: Optional[str] = None) -> WandbLogger:
        """
        Creates a new logger.

        :return: The logger.
        """
        wandb.init(entity=entity, project=project, settings=wandb.Settings(start_method='fork'))
        return cls()

    def process_py_mapper_example_queues(self, epoch: int) -> None:
        """
        Processes the example queues, logging the items to wandb.
        """
        for example_queue_name, example_queue in self.example_queues.items():
            while True:
                try:
                    queue_item = example_queue.get(block=False)
                    pass
                    if isinstance(queue_item, WandbLoggable):
                        queue_item.log(example_queue_name, epoch)
                    else:
                        raise ValueError(f"{queue_item} is not a handled logger type.")
                except queue.Empty:
                    break

    def request_examples_from_py_mapper_processes(self) -> None:
        """
        Sends requests for examples to the other processes.
        """
        for request_queue_name, request_queue in self.request_queues.items():
            request_queue.put(ExampleRequest())

    def create_callback(self) -> WandbLoggerCallback:
        """
        Creates a callback for the fit loop to call the logger methods.

        :return: The callback.
        """
        return WandbLoggerCallback(self)

    def create_request_queue_for_collection(self, name: str) -> multiprocess.Queue:
        """
        Creates a queue to send requests for examples on.

        :param name: The name of the queue.
        :return: The queue.
        """
        assert name not in self.request_queues.keys()
        manager = multiprocess.Manager()
        queue_ = manager.Queue()
        self.request_queues[name] = queue_
        return queue_

    def create_example_queue_for_collection(self, name: str) -> multiprocess.Queue:
        """
        Creates a queue to receive examples on.

        :param name: The name of the queue.
        :return: The queue.
        """
        assert name not in self.example_queues.keys()
        manager = multiprocess.Manager()
        queue_ = manager.Queue()
        self.example_queues[name] = queue_
        return queue_

    @staticmethod
    def should_produce_example(request_queue: multiprocess.Queue) -> bool:
        """
        Checks a request queue to see if an example has been requested.

        :param request_queue: The request queue.
        :return: Whether an example was requested on the queue.
        """
        try:
            request_queue.get(block=False)
            return True
        except queue.Empty:
            return False

    @staticmethod
    def submit_loggable(example_queue: multiprocess.Queue, loggable: WandbLoggable) -> None:
        """
        Submits a loggable to a request queue to be logged by the main process.

        :param example_queue: The queue to submit the loggable to.
        :param loggable: The loggable to be submitted.
        """
        while True:  # This loop should not be required, but it seems there is a bug in multiprocessing.
            try:  # https://stackoverflow.com/q/29277150/1191087
                example_queue.put(loggable)
                break
            except TypeError:
                continue

    @staticmethod
    def is_power(number: int, base: int) -> bool:
        """
        Checks if number is a power of the given base.

        :param number: The number.
        :param base: The base.
        :return: Whether the number is a power of the base.
        """
        if base in {0, 1}:
            return number == base
        power = int(math.log(number, base) + 0.5)
        return base ** power == number



from pathlib import Path
from typing import List, Optional

import numpy as np
import stringcase
import torch
import wandb
from torch.nn import BCELoss, Module
from torch.optim import AdamW
from torch.utils.data import DataLoader
from torchmetrics.classification import BinaryAccuracy

from qusi.light_curve_dataset import LightCurveDataset, InterleavedDataset
from qusi.train_hyperparameter_configuration import TrainHyperparameterConfiguration
from qusi.train_logging_configuration import TrainLoggingConfiguration
from qusi.wandb_liaison import wandb_init, wandb_log, wandb_commit


def train_session(train_datasets: List[LightCurveDataset],
                  validation_datasets: List[LightCurveDataset], model: Module,
                  hyperparameter_configuration: Optional[TrainHyperparameterConfiguration]):
    if hyperparameter_configuration is None:
        hyperparameter_configuration = TrainHyperparameterConfiguration.new()
    logging_configuration = TrainLoggingConfiguration.new()
    wandb_init(process_rank=0, project=logging_configuration.wandb_project, entity=logging_configuration.wandb_entity,
               settings=wandb.Settings(start_method='fork'))
    sessions_directory = Path('sessions')
    sessions_directory.mkdir(exist_ok=True)
    train_dataset = InterleavedDataset.new(*train_datasets)
    torch.multiprocessing.set_start_method('spawn')
    debug = False
    if debug:
        workers_per_dataloader = 0
        prefetch_factor = None
        persistent_workers = False
    else:
        workers_per_dataloader = 10
        prefetch_factor = 10
        persistent_workers = True
    train_dataloader = DataLoader(train_dataset, batch_size=hyperparameter_configuration.batch_size, pin_memory=True,
                                  persistent_workers=persistent_workers, prefetch_factor=prefetch_factor,
                                  num_workers=workers_per_dataloader)
    validation_dataloaders: List[DataLoader] = []
    for validation_dataset in validation_datasets:
        validation_dataloaders.append(DataLoader(validation_dataset, batch_size=hyperparameter_configuration.batch_size,
                                                 pin_memory=True, persistent_workers=persistent_workers,
                                                 prefetch_factor=prefetch_factor, num_workers=workers_per_dataloader))
    if torch.cuda.is_available() and not debug:
        device = torch.device('cuda')
    else:
        device = torch.device('cpu')
    model = model.to(device, non_blocking=True)
    loss_function = BCELoss().to(device, non_blocking=True)
    metric_functions = [BinaryAccuracy()]
    optimizer = AdamW(model.parameters())
    metric_functions_on_device: List[Module] = []
    for metric_function in metric_functions:
        metric_functions_on_device.append(metric_function.to(device, non_blocking=True))
    metric_functions = metric_functions_on_device
    for cycle_index in range(hyperparameter_configuration.cycles):
        train_phase(dataloader=train_dataloader, model=model, loss_function=loss_function,
                    metric_functions=metric_functions, optimizer=optimizer,
                    steps=hyperparameter_configuration.train_steps_per_cycle, device=device)
        for validation_dataloader in validation_dataloaders:
            validation_phase(dataloader=validation_dataloader, model=model, loss_function=loss_function,
                             metric_functions=metric_functions,
                             steps=hyperparameter_configuration.validation_steps_per_cycle, device=device)
        save_model(model, suffix='latest_model', process_rank=0)
        wandb_commit(process_rank=0)


def train_phase(dataloader, model, loss_function, metric_functions: List[Module], optimizer, steps, device):
    model.train()
    total_loss = 0
    metric_totals = np.zeros(shape=[len(metric_functions)])
    for batch_index, (input_features, targets) in enumerate(dataloader):
        # Compute prediction and loss
        # TODO: The conversion to float32 probably shouldn't be here, but the default collate_fn seems to be converting
        #  to float64. Probably should override the default collate.
        targets = targets.to(torch.float32).to(device, non_blocking=True)
        input_features = input_features.to(torch.float32).to(device, non_blocking=True)
        predicted_targets = model(input_features)
        loss = loss_function(predicted_targets, targets)

        # Backpropagation
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        loss, current = loss.to(device, non_blocking=True).item(), (batch_index + 1) * len(input_features)
        total_loss += loss
        for metric_function_index, metric_function in enumerate(metric_functions):
            batch_metric_value = metric_function(predicted_targets.to(device, non_blocking=True),
                                                 targets).item()
            metric_totals[metric_function_index] += batch_metric_value
        if batch_index % 10 == 0:
            print(f"loss: {loss:>7f}  [{current:>5d}/{steps * len(input_features):>5d}]", flush=True)
        if batch_index + 1 >= steps:
            break
    wandb_log('loss', total_loss / steps, process_rank=0)
    cycle_metric_values = metric_totals / steps
    for metric_function_index, metric_function in enumerate(metric_functions):
        wandb_log(f'{get_metric_name(metric_function)}', cycle_metric_values[metric_function_index],
                  process_rank=0)


def get_metric_name(metric_function):
    metric_name = type(metric_function).__name__
    metric_name = stringcase.snakecase(metric_name)
    metric_name = metric_name.replace('_metric', '').replace('_loss', '')
    return metric_name


def validation_phase(dataloader, model, loss_function, metric_functions: List[Module], steps, device):
    model.eval()
    validation_loss = 0
    metric_totals = np.zeros(shape=[len(metric_functions)])

    with torch.no_grad():
        for batch, (input_features, targets) in enumerate(dataloader):
            targets = targets.to(torch.float32).to(device, non_blocking=True)
            input_features = input_features.to(torch.float32).to(device, non_blocking=True)
            predicted_targets = model(input_features)
            validation_loss += loss_function(predicted_targets, targets).to(device, non_blocking=True).item()
            for metric_function_index, metric_function in enumerate(metric_functions):
                batch_metric_value = metric_function(predicted_targets.to(device, non_blocking=True),
                                                     targets).item()
                metric_totals[metric_function_index] += batch_metric_value
            if batch + 1 >= steps:
                break

    validation_loss /= steps
    print(f"Validation Error: \nAvg loss: {validation_loss:>8f} \n")
    wandb_log('val_loss', validation_loss, process_rank=0)
    cycle_metric_values = metric_totals / steps
    for metric_function_index, metric_function in enumerate(metric_functions):
        wandb_log(f'val_{get_metric_name(metric_function)}', cycle_metric_values[metric_function_index],
                  process_rank=0)


def save_model(model: Module, suffix: str, process_rank: int):
    if process_rank == 0:
        model_name = wandb.run.name
        if model_name == '':
            model_name = wandb.run.id
        torch.save(model.state_dict(), Path(f'sessions/{model_name}_{suffix}.pt'))



import numpy as np
from pathlib import Path
from typing import Iterable

from ramjet.photometric_database.light_curve_collection import LightCurveCollection


class SimpleLightCurveCollection(LightCurveCollection):
    """
    A simple positive and negative directory based light curve collection.
    """
    def __init__(self, collection_directory: Path = Path('data/simple_test_dataset')):
        super().__init__()
        self.collection_directory = collection_directory

    def get_paths(self) -> Iterable[Path]:
        """
        Gets the paths for the light curves in the collection.

        :return: An iterable of the light curve paths.
        """
        return self.collection_directory.glob('**/*.npz')

    def load_times_and_fluxes_from_path(self, path: Path) -> (np.ndarray, np.ndarray):
        """
        Loads the times and fluxes from a given light curve path.

        :param path: The path to the light curve file.
        :return: The times and the fluxes of the light curve.
        """
        contents = np.load(str(path))
        times = contents['times']
        fluxes = contents['fluxes']
        return times, fluxes



"""
Code for a light curve collection of the TESS two minute cadence data.
"""
import numpy as np
from pathlib import Path
from typing import Iterable, Union, List

from peewee import Select

from ramjet.data_interface.metadatabase import MetadatabaseModel
from ramjet.data_interface.tess_data_interface import TessDataInterface, TessFluxType
from ramjet.data_interface.tess_target_metadata_manager import TessTargetMetadata
from ramjet.data_interface.tess_two_minute_cadence_light_curve_metadata_manager import \
    TessTwoMinuteCadenceLightCurveMetadataManger, TessTwoMinuteCadenceLightCurveMetadata
from ramjet.photometric_database.sql_metadata_light_curve_collection import SqlMetadataLightCurveCollection


class TessTwoMinuteCadenceLightCurveCollection(SqlMetadataLightCurveCollection):
    """
    A light curve collection of the TESS two minute cadence data.
    """
    tess_data_interface = TessDataInterface()
    tess_two_minute_cadence_light_curve_metadata_manger = TessTwoMinuteCadenceLightCurveMetadataManger()

    def __init__(self, dataset_splits: Union[List[int], None] = None, flux_type: TessFluxType = TessFluxType.PDCSAP):
        super().__init__()
        self.data_directory: Path = Path('data/tess_two_minute_cadence_light_curves')
        self.label = 0
        self.dataset_splits: Union[List[int], None] = dataset_splits
        self.flux_type: TessFluxType = flux_type

    def get_sql_query(self) -> Select:
        """
        Gets the SQL query for the database models for the light curve collection.

        :return: The SQL query.
        """
        query = TessTwoMinuteCadenceLightCurveMetadata().select(TessTwoMinuteCadenceLightCurveMetadata.path)
        if self.dataset_splits is not None:
            query = query.where(TessTwoMinuteCadenceLightCurveMetadata.dataset_split.in_(self.dataset_splits))
        return query

    def get_path_from_model(self, model: MetadatabaseModel) -> Path:
        """
        Gets the light curve path from the SQL database model.

        :return: The path to the light curve.
        """
        return Path(self.tess_two_minute_cadence_light_curve_metadata_manger.
                    light_curve_root_directory_path.joinpath(model.path))

    def load_times_and_fluxes_from_path(self, path: Path) -> (np.ndarray, np.ndarray):
        """
        Loads the times and fluxes from a given light curve path.

        :param path: The path to the light curve file.
        :return: The times and the fluxes of the light curve.
        """
        fluxes, times = load_fluxes_and_times_from_fits_file(path, self.flux_type)
        return times, fluxes

    def load_times_and_magnifications_from_path(self, path: Path) -> (np.ndarray, np.ndarray):
        """
        Loads the times and magnifications from a given path as an injectable signal.

        :param path: The path to the light curve/signal file.
        :return: The times and the magnifications of the light curve/signal.
        """
        fluxes, times = load_fluxes_and_times_from_fits_file(path, self.flux_type)
        magnifications, times = self.generate_synthetic_signal_from_real_data(fluxes, times)
        return times, magnifications

    def download(self):
        """
        Downloads the light curve collection.
        """
        download_two_minute_cadence_light_curves(self.data_directory)


class TessTwoMinuteCadenceTargetDatasetSplitLightCurveCollection(TessTwoMinuteCadenceLightCurveCollection):
    """
    A light curve collection of the TESS two minute cadence data with light curves from the same target in the same
    dataset split.
    """
    def get_sql_query(self) -> Select:
        """
        Gets the SQL query for the database models for the light curve collection.

        :return: The SQL query.
        """
        query = TessTwoMinuteCadenceLightCurveMetadata().select(TessTwoMinuteCadenceLightCurveMetadata.path)
        query = query.join(TessTargetMetadata,
                           on=TessTwoMinuteCadenceLightCurveMetadata.tic_id == TessTargetMetadata.tic_id)
        if self.dataset_splits is not None:
            query = query.where(TessTargetMetadata.dataset_split.in_(self.dataset_splits))
        return query



"""
A model for the database entity of a TESS planet disposition.
"""
from enum import Enum
from typing import List

from peewee import ForeignKeyField, CharField, TextField, AutoField

from ramjet.database.base_model import BaseModel
from ramjet.database.tess_target import TessTarget
from ramjet.database.tess_transiter import TessTransiter


class Disposition(Enum):
    """
    An enum to represent the possible planet dispositions.
    """
    PASS = 'Pass'
    CONDITIONAL = 'Conditional'
    AMBIGUOUS = 'Ambiguous'
    UNLIKELY = 'Unlikely'
    FAIL = 'Fail'
    REPROCESSING_REQUIRED = 'Reprocessing required'
    KNOWN = 'Known'


class Source(Enum):
    """
    An enum to represent the possible sources of planet dispositions.
    """
    GREG_OLMSCHENK = 'Greg Olmschenk'
    GSFC_GROUP = 'GSFC group'


class TessPlanetDisposition(BaseModel):
    """
    A database model for the database entity of a TESS planet disposition.
    """
    id = AutoField()
    transiter: TessTransiter = ForeignKeyField(TessTransiter)
    disposition = CharField(choices=Disposition)
    source = CharField(choices=Source)
    notes = TextField(null=True)

    class Meta:
        """Schema meta data for the model."""
        indexes = (
            (('source', 'transiter'), True),
        )

    @staticmethod
    def get_tic_ids_of_passing_vetted_transiting_planet_candidates() -> List[int]:
        """
        Gets the TIC IDs of candidates which have passed vetting of having planet transits.

        :return: The list of candidate TIC IDs.
        """
        candidate_tic_id_query = TessTarget.select(TessTarget.tic_id).join(TessTransiter).where(
            TessTransiter.id.in_(
                TessPlanetDisposition.select(TessPlanetDisposition.transiter).where(
                    TessPlanetDisposition.disposition == Disposition.PASS.value))
            &
            TessTransiter.id.not_in(
                TessPlanetDisposition.select(TessPlanetDisposition.transiter).where(
                    TessPlanetDisposition.disposition == Disposition.FAIL.value))
        ).order_by(TessTarget.tic_id)
        return [candidate.tic_id for candidate in candidate_tic_id_query]